"""
csv streamer
"""
import csv

from flume.adapters.streamers.base import Streamer
from flume import Point


class CSV(Streamer):

    def __init__(self,
                 headers=True,
                 delimiter=',',
                 ignore_whitespace=True):
        self.writer = None
        self.headers = headers
        self.delimiter = delimiter
        self.ignore_whitespace = ignore_whitespace

    def read(self, stream):
        reader = csv.DictReader(stream.readlines(),
                                delimiter=self.delimiter,
                                skipinitialspace=self.ignore_whitespace)

        for point in reader:
            # XXX: buffer N points before pushing out ? 
            yield Point(**point)

    def write(self, stream, points):
        for point in points:
            point = point.json()

            if self.writer is None:
                fieldnames = point.keys()
                self.writer = csv.DictWriter(stream, fieldnames=fieldnames)
                if self.headers:
                    self.writer.writeheader()

            self.writer.writerow(point)



"""
csv streamer unittests
"""
import unittest

from robber import expect

from flume import Point
from flume.adapters.streamers import get_streamer
from test.unit.util import FakeIO


class CSVStreamerTest(unittest.TestCase):

    def test_csv_streamer_can_read_an_empty_stream(self):
        streamer = get_streamer('csv')
        stream = FakeIO('')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([])

    def test_csv_streamer_can_read_a_single_point(self):
        streamer = get_streamer('csv')
        stream = FakeIO('time\n2016-01-01T00:00:00.000Z\n')

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z')
        ])

    def test_csv_streamer_can_read_multiple_points(self):
        streamer = get_streamer('csv')
        stream = FakeIO('time,foo\n' +
                        '2016-01-01T00:00:00.000Z,1\n' +
                        '2016-01-01T00:00:01.000Z,2\n' +
                        '2016-01-01T00:00:02.000Z,3')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z', foo='1'),
            Point(time='2016-01-01T00:00:01.000Z', foo='2'),
            Point(time='2016-01-01T00:00:02.000Z', foo='3')
        ])

    def test_csv_streamer_can_handle_blank_lines(self):
        streamer = get_streamer('csv')
        stream = FakeIO('time,foo\n' +
                        '2016-01-01T00:00:00.000Z,1\n' +
                        '\n' +
                        '2016-01-01T00:00:01.000Z,2\n')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z', foo='1'),
            Point(time='2016-01-01T00:00:01.000Z', foo='2')
        ])

    def test_csv_streamer_can_write_an_empty_stream(self):
        streamer = get_streamer('csv')
        stream = FakeIO('')

        streamer.write(stream, [])
        stream = FakeIO(stream.getvalue())

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([])

    def test_csv_streamer_can_write_a_single_point(self):
        streamer = get_streamer('csv')
        stream = FakeIO('')

        streamer.write(stream, [
            Point(time='2016-01-01T00:00:00.000Z')
        ])
        stream = FakeIO(stream.getvalue())

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z')
        ])

    def test_csv_streamer_can_write_multiple_points(self):
        streamer = get_streamer('csv')
        stream = FakeIO('')

        streamer.write(stream, [
            Point(time='2016-01-01T00:00:00.000Z', foo=1),
            Point(time='2016-01-01T00:00:01.000Z', foo=2),
            Point(time='2016-01-01T00:00:02.000Z', foo=3)
        ])
        stream = FakeIO(stream.getvalue())

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z', foo='1'),
            Point(time='2016-01-01T00:00:01.000Z', foo='2'),
            Point(time='2016-01-01T00:00:02.000Z', foo='3')
        ])



"""
http adapter
"""
import requests
import requests_cache

from flume import logger, moment
from flume.adapters.adapter import adapter
from flume.adapters import streamers
from flume.exceptions import FlumineException
from flume.point import Point


class RequestsStream(object):

    def __init__(self, response):
        self.response = response

    def read(self, size=1024):
        return self.response.iter_content(chunk_size=size)

    def readlines(self):
        tail = ''

        for data in self.read():
            data = tail + data
            lines = data.split('\n')

            head = lines[:-1]
            tail = lines[-1]

            for line in head:
                yield line

        for line in tail:
            yield line


class http(adapter):
    """
    http adapter
    """

    name = 'http'


    def __init__(self,
                 url=None,
                 method='GET',
                 headers=None,
                 follow_link=True,
                 format=None,
                 cache=None,
                 array=True,
                 status=[200],
                 **kwargs):
        self.url = url
        self.method = method
        self.headers = headers
        self.follow_link = follow_link
        self.cache = cache
        self.array = array

        if not isinstance(status, list):
            status = [status]

        self.status = status

        if format is not None:
            self.streamer = streamers.get_streamer(format, **kwargs)

        else:
            self.streamer = None

    def read(self):
        """
        read points by issuing an HTTP request and pushing the response as
        individual points into the flume pipeline
        """
        url = self.url

        while True:
            if self.cache is not None:
                requests_cache.install_cache(self.cache)

            def verify_response(response):
                if response.status_code not in self.status:
                    raise FlumineException('%s: %s' % (response.status_code, response.text))

            if self.streamer is not None:
                response = requests.request(self.method,
                                            url,
                                            headers=self.headers,
                                            stream=True)
                verify_response(response)
                data = self.streamer.read(RequestsStream(response))

            else:
                response = requests.request(self.method, url, headers=self.headers)
                verify_response(response)
                data = response.json()

            points = []
            for point in data:
                # convert to a flume Point
                point = Point(**point)
                points.append(point)

            yield points

            if not self.follow_link:
                break

            if 'link' in response.headers and 'next' in response.links:
                url = response.links['next']['url']

            else:
                break

    def __push(self, payload):
        response = requests.request(self.method,
                                    self.url,
                                    headers=self.headers,
                                    json=payload)

        if response.status_code not in self.status:
            raise FlumineException('received bad response with %d: %s' %
                                   (response.status_code, response.text))

    def write(self, points):
        if self.array:
            payload = []

            for point in points:
                payload.append(point.json())

            self.__push(payload)

        else:
            for point in points:
                self.__push(point.json())



"""
iterate reducer
"""
from flume import reducer


class iterate(reducer):

    def __init__(self, values):
        self.values = values
        self.index = -1

    def update(self, point):
        self.index += 1

    def result(self):
        return self.values[self.index % len(self.values)]

    def reset(self):
        self.index = 0



"""
github issues API v3

  https://developer.github.com/v3/issues/

Not everything has been covered here as this is more of an example of how to
create a third party integration that makes getting external data into flume
easy.
"""

from flume.sources import read
from flume.thirdparty.github.util import github_url


def user(oauth=None,
         **kwargs):
    headers = {
        'Authorization': 'token %s' % oauth
    }

    return read('http',
                url=github_url('user/issues', **kwargs),
                time='created_at',
                headers=headers)

def orgs(org_name,
         oauth=None,
         **kwargs):
    headers = {
        'Authorization': 'token %s' % oauth
    }

    return read('http',
                url=github_url('orgs/%s/issues' % org_name, **kwargs),
                time='created_at',
                headers=headers)

def repo(owner,
         repo_name,
         oauth=None,
         **kwargs):
    headers = {}

    if oauth is not None:
        headers['Authorization'] = 'token %s' % oauth

    return read('http',
                url=github_url('repos/%s/%s/issues' %
                                (owner, repo_name),
                                **kwargs),
                time='created_at',
                headers=headers)

def issue(owner,
          repo_name,
          issue_number,
          oauth=None,
          **kwargs):
    headers = {}

    if oauth is not None:
        headers['Authorization'] = 'token %s' % oauth

    return read('http',
                url=github_url('repos/%s/%s/issues/%s' %
                                (owner, repo_name, issue_number),
                                **kwargs),
                time='created_at',
                headers=headers)

def comments(owner,
             repo_name,
             issue_number,
             oauth=None,
             **kwargs):
    headers = {}

    if oauth is not None:
        headers['Authorization'] = 'token %s' % oauth

    return read('http',
                url=github_url('repos/%s/%s/issues/%s/comments' %
                                (owner, repo_name, issue_number),
                                **kwargs),
                time='created_at',
                headers=headers)



"""
linechart sink
"""

from flume import logger
from flume import sink
from flume.sinks.views import get_view


class linechart(sink):

    name = 'linechart'

    def __init__(self,
                 view,
                 series='name',
                 yvalue='value',
                 xvalue='time',
                 **kwargs):
        sink.__init__(self)
        self.series = series
        self.xvalue = xvalue
        self.yvalue = yvalue
        self.view = get_view(view, **kwargs)

    def loop(self):
        series = {}

        while self.running:
            points = self.pull()

            for point in points:
                if self.series not in point:
                    logger.warn('unable to find series field "%s" in point' % self.series)
                    name = 'unknown'
                else:
                    name = point[self.series]

                if name not in series:
                    series[name] = []

                series[name].append((point[self.xvalue],
                                     point[self.yvalue]))

            self.view.render(self.name, series)



"""
memory sink
"""

from flume import sink
from flume.exceptions import FlumineException

class memory(sink):

    name = 'memory'

    def __init__(self, results):
        sink.__init__(self, results)

        if not isinstance(results, list):
            raise FlumineException('results must a list')

        self.results = results

    def loop(self):
        while self.running:
            points = self.pull()

            for point in points:
                self.results.append(point.json())



"""
json streamer
"""
import ijson
import json

from flume.adapters.streamers.base import Streamer
from flume.point import Point

class JSON(Streamer):

    def __init__(self, pretty=False):
        self.pretty = pretty
        self.first_point = True
        self.start = False
        self.writing = False

    def read(self, stream):
        for point in ijson.items(stream, 'item'):
            yield Point(**point)

    def write(self, stream, points):
        self.writing = True

        if not self.start:
            stream.write('[\n')
            self.start = True

        for point in points:
            if not self.first_point:
                stream.write(',\n')

            else:
                self.first_point = False

            if self.pretty:
                # you can sort the keys in pretty jsonl format but indenting
                # would mean breaking the jsonl format
                string = json.dumps(point.json(), sort_keys=True, indent=4)

            else:
                string = json.dumps(point.json())

            stream.write('%s' % string)

    def eof(self, stream):
        if self.writing:
            stream.write('\n]')



"""
base adapter unittests
"""

import unittest

from flume.adapters import register_adapter, get_adapter
from flume.adapters import adapter
from flume.exceptions import FlumineException

from robber import expect


class AdaptersTest(unittest.TestCase):

    def test_register_non_unique_adapter_fails(self):
        class TestAdapter(adapter):
            name = 'non-unique'

        register_adapter(TestAdapter)
        try:
            register_adapter(TestAdapter)
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('"non-unique" adapter already registered')

    def test_register_and_get_adapter_by_name(self):
        class TestAdapter(adapter):
            name = 'another'

            def read(self, stream):
                pass

        register_adapter(TestAdapter)
        result = get_adapter('another', 'read')
        expect(result).to.eq(TestAdapter)

    def test_get_adapter_of_inexistent_adapter_fails(self):
        try:
            get_adapter('non-existent', 'read')
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('"non-existent" adapter not registered')

    def test_register_and_get_adapter_by_name_with_missing_operation_fails(self):
        class TestAdapter(adapter):
            name = 'not-writable'

        register_adapter(TestAdapter)
        try:
            get_adapter('not-writable', 'write')
        except FlumineException as exception:
            expect(exception.message).to.contain('"not-writable" adapter does not support write')



"""
head processor
"""
from flume import node


class head(node):

    name = 'head'

    def __init__(self, howmany):
        node.__init__(self)
        self.howmany = howmany

    def loop(self):
        points = []
        done = False

        while self.running:
            points += self.pull()

            if len(points) < self.howmany:
                continue

            if not done:
                done = True
                self.push(points[:self.howmany])

        if not done:
            done = True
            self.push(points[:self.howmany])



"""
tail processor
"""
from flume import node


class tail(node):

    name = 'tail'

    def __init__(self, howmany):
        node.__init__(self)
        self.howmany = howmany

    def loop(self):
        points = []

        while self.running:
            points += self.pull()
            points = points[-self.howmany:]

        self.push(points[-self.howmany:])



"""
put processor
"""
from flume import node, reducer
from flume.util import is_string

class put(node):

    name = 'put'

    def __init__(self, **fields):
        node.__init__(self, **fields)
        self.fields = fields

    def loop(self):
        while self.running:
            points = self.pull()

            for point in points:
                # call update every time through here
                for (key, value) in self.fields.items():
                    if isinstance(value, reducer):
                        value.update(point)
                        point[key] = value.result()

                    else:
                        if is_string(value):
                            point[key] = value.format(**point.json())

                        else:
                            point[key] = value

            self.push(points)



"""
funcr reducer - transforms any existing function into a reducer that can be used
                in the flume pipeline
"""

from flume import reducer

def funcr(func):

    class wrapper_reducer(reducer):
        def __init__(self, *fieldnames):
            self.fieldnames = fieldnames
            self.value = None

        def update(self, point):
            self.value = func(*[point[fieldname] for fieldname in self.fieldnames])

        def result(self):
            return self.value

        def reset(self):
            self.value = None

    return wrapper_reducer



"""
flume data point class
"""
import json
from datetime import datetime

from flume import moment
from dici import dici


def dmap(dictionary, func):
    result = {}

    for (key, value) in dictionary.items():
        if isinstance(value, dict):
            value = dmap(value, func)

        result[key] = func(value)

    return result

class Point(dici):
    """
    This class represents a data point within the flume pipeline
    """

    def __init__(self, **fields):
        fields = {} if fields is None else fields
        dici.__init__(self, **fields)

        if '__meta__' not in self:
            self['__meta__'] = {'flume_path': ''}

    def hasfield(self, dot_notation):
        """
        returns True if the point has the field expressed with dot notation
        """
        return self.lookup(dot_notation) is not None

    def lookup(self, dot_notation):
        """
        lookup the value of a field specified with the dot notation
        """
        try:
            return self.__getattr__(dot_notation)
        except AttributeError:
            pass
        except KeyError:
            pass

        return None

    def update(self, other):
        """
        update only updates the points fields that aren't within the __meta__ object
        """
        for key in other.keys():
            if key != '__meta__':
                self[key] = other[key]

    def json(self):
        """
        return the JSON serializable representation of this object where all
        of the complex members have been converted into strings
        """
        def handle_dates(value):
            if isinstance(value, datetime):
                return moment.datetime_to_iso8601(value)
            else:
                return value

        result = dmap(self, handle_dates)
        del result['__meta__']
        return result

    def __str__(self):
        return json.dumps(self.json(), sort_keys=True)



from flume import reducer


class minimum(reducer):

    def __init__(self, fieldname):
        self.fieldname = fieldname
        self.min = None

    def update(self, point):
        if self.min:
            if point[self.fieldname] < self.min:
                self.min = point[self.fieldname]
        else:
            self.min = point[self.fieldname]

    def result(self):
        return self.min

    def reset(self):
        self.min = None



"""
remove processor
"""
from flume import node


class remove(node):
    """
    # remove

    The remove processor is used to to remove unwanted fields from the stream by
    specifying the ones you want to remove

    usage: ... | remove(field,...,fieldN) | ...

    """

    name = 'remove'

    def __init__(self, *fieldnames):
        node.__init__(self, *fieldnames)
        self.fieldnames = fieldnames

    def loop(self):
        while self.running:
            points = self.pull()

            for point in points:
                for fieldname in self.fieldnames:
                    if point.hasfield(fieldname):
                        del point[fieldname]

            self.push(points)



import unittest

from robber import expect
from flume import moment


class MomentTest(unittest.TestCase):

    def test_parsing_iso8601_date(self):
        date1 = moment.date('2016-01-01')
        date2 = moment.date('2016-01-01T00:00:00.000Z')
        expect(date1).to.eq(date2)

    def test_parsing_a_relative_day_ago_date(self):
        date1 = moment.date('1 day ago')
        date2 = moment.date('24 hours ago')
        expect(date1).to.eq(date2)

    def test_parsing_a_relative_year_ago_date(self):
        date1 = moment.date('1 year ago')
        date2 = moment.date('12 months ago')
        expect(date1).to.eq(date2)



"""
sort processor
"""
import operator

from flume import node
from flume.exceptions import FlumineException


class sort(node):

    name = 'sort'

    def __init__(self, *fieldnames, **kwargs):
        node.__init__(self, *fieldnames)
        self.fieldnames = fieldnames

        if 'limit' in kwargs:
            self.limit = kwargs['limit']

        else:
            self.limit = 100000

        if 'order' in kwargs:
            order = kwargs['order']

            if order not in ['asc', 'desc']:
                raise FlumineException('order can be "asc" or "desc", not "%s"' % order)

            self.order = order

        else:
            self.order = 'asc'

    def loop(self):
        result = []

        while self.running:
            points = self.pull()

            if len(result) > self.limit:
                raise FlumineException('sort buffer overflown, limit is %s, buffering %s' %
                                       (self.limit, len(result)))

            result += points

        if len(self.fieldnames) > 0:
            result = sorted(result,
                            key=operator.attrgetter(*self.fieldnames),
                            reverse=(self.order == 'desc'))

            # XXX: we could remove the field while we sort to make this a bit
            #      more efficient
            for point in result:
                if 'time' in point:
                    del point['time']

        self.push(result)



"""
base adapter class
"""

class adapter(object):
    """
    Adapter base class used to define the required adapter methods
    """

    def read(self):
        """
        return a generator of points from this adapter
        """
        pass

    def write(self, points):
        """
        write the points provided through this adapter
        """
        pass

    def eof(self):
        """
        finalize any necessary things in the adapter (used only on write)
        """
        pass



"""
remove proc unittests
"""
import unittest

from robber import expect
from flume import *


class SortTest(unittest.TestCase):

    def test_sort_on_invalid_order(self):
        try:
            (
                emit(limit=10, start='2016-01-01', every='1s')
                | put(count=count())
                | sort('count', order='bananas')
            ).execute()

            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('order can be "asc" or "desc", not "bananas"')

    def test_sort_limits_exceeded(self):
        try:
            (
                emit(limit=10, start='2016-01-01', every='1s')
                | put(count=count())
                | sort('count', limit=5)
            ).execute()

            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('sort buffer overflown, limit is 5, buffering 6')

    def test_sort_with_no_fields_is_idempotent(self):
        results = []
        (
            emit(limit=5, start='2016-01-01', every='1s')
            | put(count=count())
            | sort()
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T00:00:01.000Z', 'count': 2},
            {'time': '2016-01-01T00:00:02.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'count': 4},
            {'time': '2016-01-01T00:00:04.000Z', 'count': 5}
        ])

    def test_sort_ascending_on_single_inorder_field(self):
        results = []
        (
            emit(limit=5, start='2016-01-01', every='1s')
            | put(count=count())
            | sort('count', order='asc')
            | memory(results)
        ).execute()

        # time should be removed since its no longer in order
        expect(results).to.eq([
            {'count': 1},
            {'count': 2},
            {'count': 3},
            {'count': 4},
            {'count': 5}
        ])

    def test_sort_descending_on_single_inorder_field(self):
        results = []
        (
            emit(limit=5, start='2016-01-01', every='1s')
            | put(count=count())
            | sort('count', order='desc')
            | memory(results)
        ).execute()

        # time should be removed since its no longer in order
        expect(results).to.eq([
            {'count': 5},
            {'count': 4},
            {'count': 3},
            {'count': 2},
            {'count': 1}
        ])

    def test_sort_ascending_on_multiple_fields(self):
        results = []
        (
            emit(limit=3, start='2016-01-01', every='1s')
            | put(count=count())
            | (
                put(foo=iterate(['a', 'b'])),
                put(foo=iterate(['b', 'a']))
            ) | sort('foo', 'count', order='asc')
            | memory(results)
        ).execute()

        # time should be removed since its no longer in order
        expect(results).to.eq([
            {'foo': 'a', 'count': 1},
            {'foo': 'a', 'count': 2},
            {'foo': 'a', 'count': 3},
            {'foo': 'b', 'count': 1},
            {'foo': 'b', 'count': 2},
            {'foo': 'b', 'count': 3}
        ])

    def test_sort_descending_on_multiple_fields(self):
        results = []
        (
            emit(limit=3, start='2016-01-01', every='1s')
            | put(count=count())
            | (
                put(foo=iterate(['a', 'b'])),
                put(foo=iterate(['b', 'a']))
            ) | sort('foo', 'count', order='desc')
            | memory(results)
        ).execute()

        # time should be removed since its no longer in order
        expect(results).to.eq([
            {'foo': 'b', 'count': 3},
            {'foo': 'b', 'count': 2},
            {'foo': 'b', 'count': 1},
            {'foo': 'a', 'count': 3},
            {'foo': 'a', 'count': 2},
            {'foo': 'a', 'count': 1}
        ])



"""
github pulls API v3

  https://developer.github.com/v3/pulls

Not everything has been covered here as this is more of an example of how to
create a third party integration that makes getting external data into flume
easy.
"""

from flume.sources import read
from flume.thirdparty.github.util import github_url


def repo(owner,
         repo_name,
         oauth=None,
         **kwargs):
    headers = {}

    if oauth is not None:
        headers['Authorization'] = 'token %s' % oauth

    return read('http',
                url=github_url('repos/%s/%s/pulls' % (owner, repo_name),
                               **kwargs),
                time='created_at',
                headers=headers)

def commits(owner,
            repo_name,
            pull_number,
            oauth=None):
    headers = {}

    if oauth is not None:
        headers['Authorization'] = 'token %s' % oauth

    return read('http',
                url=github_url('repos/%s/%s/pulls/%s/commits' %
                               (owner, repo_name, pull_number)),
                time='created_at',
                headers=headers)



"""
count reducer
"""
from flume import reducer


class count(reducer):

    def __init__(self):
        self.count = 0

    def update(self, point):
        self.count += 1

    def result(self):
        return self.count

    def reset(self):
        self.count = 0



"""
head proc unittests
"""
import unittest

from robber import expect
from flume import *


class HeadTest(unittest.TestCase):

    def test_head_on_empty_stream_works(self):
        results = []
        (
            emit(points=[])
            | head(3)
            | memory(results)
        ).execute()
        expect(results).to.eq([])

    def test_head_on_stream_with_less_than_N_elements(self):
        results = []
        (
            emit(points=[
                {'foo': 1},
                {'foo': 2},
                {'foo': 3}
            ])
            | head(5)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1},
            {'foo': 2},
            {'foo': 3}
        ])

    def test_head_on_stream_with_more_than_N_elements(self):
        results = []
        (
            emit(points=[
                {'foo': 1},
                {'foo': 2},
                {'foo': 3},
                {'foo': 4},
                {'foo': 5}
            ])
            | head(3)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1},
            {'foo': 2},
            {'foo': 3}
        ])



"""
barchart unit tests
"""
import unittest
import mock

from flume import *
from flume import moment
from flume.adapters.adapter import adapter

from robber import expect


class BarchartTest(unittest.TestCase):

    @mock.patch('flume.sinks.views.gnuplot.Gnuplot.render')
    def test_can_handle_an_empty(self, mock_render):
        (
            emit(points=[])
            | barchart('gnuplot')
        ).execute()
        expect(mock_render.call_args).to.eq(mock.call('barchart', {}))

    @mock.patch('flume.sinks.views.gnuplot.Gnuplot.render')
    def test_can_handle_a_single_series(self, mock_render):
        (
            emit(points=[
                {'time': '2016-01-01T00:00:01.000Z', 'name': 'series1', 'value': 1},
                {'time': '2016-01-01T00:00:02.000Z', 'name': 'series1', 'value': 2},
                {'time': '2016-01-01T00:00:03.000Z', 'name': 'series1', 'value': 3}
            ])
            | barchart('gnuplot')
        ).execute()
        expect(mock_render.call_args).to.eq(mock.call('barchart', {
            'series1': [
                (moment.date('2016-01-01T00:00:01.000Z'), 1),
                (moment.date('2016-01-01T00:00:02.000Z'), 2),
                (moment.date('2016-01-01T00:00:03.000Z'), 3)
            ]
        }))

    @mock.patch('flume.sinks.views.gnuplot.Gnuplot.render')
    def test_can_handle_multiple_series(self, mock_render):
        (
            emit(points=[
                {'time': '2016-01-01T00:00:01.000Z', 'name': 'series1', 'value': 1},
                {'time': '2016-01-01T00:00:02.000Z', 'name': 'series2', 'value': 2},
                {'time': '2016-01-01T00:00:03.000Z', 'name': 'series1', 'value': 3},
                {'time': '2016-01-01T00:00:04.000Z', 'name': 'series2', 'value': 4},
                {'time': '2016-01-01T00:00:05.000Z', 'name': 'series1', 'value': 5},
                {'time': '2016-01-01T00:00:06.000Z', 'name': 'series2', 'value': 6}
            ])
            | barchart('gnuplot')
        ).execute()
        expect(mock_render.call_args).to.eq(mock.call('barchart', {
            'series1': [
                (moment.date('2016-01-01T00:00:01.000Z'), 1),
                (moment.date('2016-01-01T00:00:03.000Z'), 3),
                (moment.date('2016-01-01T00:00:05.000Z'), 5)
            ],
            'series2': [
                (moment.date('2016-01-01T00:00:02.000Z'), 2),
                (moment.date('2016-01-01T00:00:04.000Z'), 4),
                (moment.date('2016-01-01T00:00:06.000Z'), 6)
           ]
        }))

    @mock.patch('flume.logger.warn')
    @mock.patch('flume.sinks.views.gnuplot.Gnuplot.render')
    def test_can_handle_a_series_with_no_series_field_and_no_other_tags(self, mock_render, mock_warn):
        (
            emit(points=[
                {'time': '2016-01-01T00:00:01.000Z', 'value': 1},
                {'time': '2016-01-01T00:00:02.000Z', 'value': 2},
                {'time': '2016-01-01T00:00:03.000Z', 'value': 3}
            ])
            | barchart('gnuplot')
        ).execute()
        expect(mock_render.call_args).to.eq(mock.call('barchart', {
            'unknown': [
                (moment.date('2016-01-01T00:00:01.000Z'), 1),
                (moment.date('2016-01-01T00:00:02.000Z'), 2),
                (moment.date('2016-01-01T00:00:03.000Z'), 3)
            ]
        }))
        expect(mock_warn.call_args_list).to.eq([
            mock.call('unable to find series field "name" in point'),
            mock.call('unable to find series field "name" in point'),
            mock.call('unable to find series field "name" in point')
        ])

    @mock.patch('flume.logger.warn')
    @mock.patch('flume.sinks.views.gnuplot.Gnuplot.render')
    def test_can_handle_a_series_with_no_series_field_but_with_other_tags(self, mock_render, mock_warn):
        (
            emit(points=[
                {'time': '2016-01-01T00:00:01.000Z', 'value': 1, 'series': 1},
                {'time': '2016-01-01T00:00:02.000Z', 'value': 2, 'series': 1},
                {'time': '2016-01-01T00:00:03.000Z', 'value': 3, 'series': 1}
            ])
            | barchart('gnuplot')
        ).execute()
        expect(mock_render.call_args).to.eq(mock.call('barchart', {
            'unknown': [
                (moment.date('2016-01-01T00:00:01.000Z'), 1),
                (moment.date('2016-01-01T00:00:02.000Z'), 2),
                (moment.date('2016-01-01T00:00:03.000Z'), 3)
            ]
        }))
        expect(mock_warn.call_args_list).to.eq([
            mock.call('unable to find series field "name" in point'),
            mock.call('unable to find series field "name" in point'),
            mock.call('unable to find series field "name" in point')
        ])



"""
flume init module

- place to define what gets exported from the flume package and how
"""

import flume.logger as logger

from flume.point import Point

from flume.adapters import adapter, \
                           elastic, \
                           http, \
                           stdio, \
                           register_adapter, \
                           get_adapter

from flume.core import node, \
                       queue, \
                       reducer, \
                       sink, \
                       register_source, \
                       register_proc, \
                       register_sink

from flume.thirdparty import *

from flume.exceptions import *

from flume.moment import *
from flume.sources import *
from flume.procs import *
from flume.procs import *
from flume.sinks import *

from flume.procs.reduce.reducers import *

"""
register built in sources, procs, sinks, etc of flume
"""

register_source(emit)
register_source(read)

register_proc(intersect)
register_proc(keep)
register_proc(put)
register_proc(head)
register_proc(tail)
register_proc(reduce)
register_proc(remove)
register_proc(reorder)
register_proc(sort)
register_proc(union)

register_sink(memory)
register_sink(write)

register_sink(barchart)
register_sink(linechart)

register_adapter(http)
register_adapter(elastic)
register_adapter(stdio)



"""
stdio adapter
"""
import codecs
import re
import sys
import zlib

from flume.adapters.adapter import adapter
from flume.adapters import streamers
from flume.exceptions import FlumineException
from flume.point import Point
from flume import util

# For things to work in Python 3
# from: http://python-notes.curiousefficiency.org/en/latest/python3/text_file_processing.html#unicode-basics
# "Approach: use the 'latin-1' encoding to map byte values directly to the
#  first 256 Unicode code points. This is the closest equivalent Python 3
#  offers to the permissive Python 2 text handling model."
_DEFAULT_ENCODING = 'latin-1'


class InputStream(object):
    """
    input stream wrapper that gives us a few things:

      * override the behavior of readlines which doesn't actually stream data
        but attempts to read the whole input and break it into individual
        lines. We want the input to stream line by line and not block the
        pipeline.

      * with strip_ansi set to True this wrapper will make sure to strip any
        ANSI sequences from the stream.

      * handles compressed streams by decompressing using the algorithm
        specified by the compression option.
    """

    def __init__(self,
                 stream,
                 strip_ansi=False,
                 compression=None):
        self.stream = stream
        self.strip_ansi = strip_ansi
        self.decompressor = None
        self.__line_buffer = ''

        if compression is not None:
            if compression == 'gzip':
                self.decompressor = zlib.decompressobj(zlib.MAX_WBITS | 16)

            elif compression == 'zlib':
                self.decompressor = zlib.decompressobj(zlib.MAX_WBITS)

            elif compression == 'deflate':
                self.decompressor = zlib.decompressobj(-zlib.MAX_WBITS)

            else:
                raise FlumineException('unsupported compression [%s]' % compression)

    def __remove_ansi(self, line):
        return re.sub(r'\x1b[^A-Za-z]*[A-Za-z]', '', line)

    def read(self, size=1024):
        data = self.stream.read(size)

        if data is None:
            data = ''

        if self.decompressor is not None:
            if util.IS_PY2:
                data = self.decompressor.decompress(data)

            else:
                data = self.decompressor.decompress(data.encode(_DEFAULT_ENCODING))
                data = data.decode(_DEFAULT_ENCODING)

        if self.strip_ansi:
            data = self.__remove_ansi(data)

        return data

    def readline(self, limit=-1):
        index = self.__line_buffer.find('\n')

        while index == -1:
            data = self.read()

            if not data:
                data = self.__line_buffer
                self.__line_buffer = ''
                return data

            self.__line_buffer += data
            index = self.__line_buffer.find('\n')

        result = self.__line_buffer[:index]
        self.__line_buffer = self.__line_buffer[index + 1:]

        return result

    def readlines(self):

        while True:
            line = self.readline()

            if not line:
                break

            if self.strip_ansi:
                yield self.__remove_ansi(line)

            else:
                yield line

class OutputStream(object):

    def __init__(self,
                 stream,
                 compression=None):
        self.stream = stream
        self.compressor = None

        if compression is not None:
            if compression == 'gzip':
                self.compressor = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                                                   zlib.DEFLATED,
                                                   zlib.MAX_WBITS | 16)

            elif compression == 'zlib':
                self.compressor = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                                                   zlib.DEFLATED,
                                                   zlib.MAX_WBITS)

            elif compression == 'deflate':
                self.compressor = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                                                   zlib.DEFLATED,
                                                   -zlib.MAX_WBITS)

            else:
                raise FlumineException('unsupported compression [%s]' % compression)

    def write(self, data):
        if self.compressor is not None:
            if util.IS_PY2:
                data = self.compressor.compress(data)

            else:
                data = self.compressor.compress(data.encode(_DEFAULT_ENCODING))
                data = data.decode(_DEFAULT_ENCODING)

        self.stream.write(data)

    def flush(self):
        if self.compressor is not None:
            if util.IS_PY2:
                data = self.compressor.flush()
                self.stream.write(data)

            else:
                data = self.compressor.flush().decode(_DEFAULT_ENCODING)
                self.stream.write(data)

        self.stream.flush()

    def close(self):
        self.flush()
        self.stream.close()

class stdio(adapter):
    """
    stdio adapter

    Used to read/write points from standard IO which includes the stdin,
    stdout and stderr, but can also be used to read/write points to a
    file on the filesystem
    """

    name = 'stdio'

    stdout = sys.stdout
    stdin = sys.stdin

    def __init__(self,
                 format='jsonl',
                 file=None,
                 append=False,
                 strip_ansi=False,
                 compression=None,
                 **kwargs):
        self.streamer = streamers.get_streamer(format, **kwargs)
        self.file = file
        self.append = append
        self.strip_ansi = strip_ansi
        self.compression = compression

        self.output = None

    def read(self):
        if self.file is None:
            input_stream = InputStream(stdio.stdin,
                                       strip_ansi=self.strip_ansi,
                                       compression=self.compression)

            for point in self.streamer.read(input_stream):
                yield [Point(**point)]

        else:
            if util.IS_PY2:
                encoding = None

            else:
                encoding = _DEFAULT_ENCODING

            with codecs.open(self.file, 'r', encoding=encoding) as stream:
                stream = InputStream(stream,
                                     strip_ansi=self.strip_ansi,
                                     compression=self.compression)

                for point in self.streamer.read(stream):
                    yield [Point(**point)]

    def write(self, points):
        if self.output is None:
            if self.file is None:
                self.output = OutputStream(stdio.stdout,
                                           compression=self.compression)
            else:
                if self.append:
                    mode = 'a'

                else:
                    mode = 'w'

                if util.IS_PY2:
                    encoding = None

                else:
                    encoding = _DEFAULT_ENCODING

                output = codecs.open(self.file, mode, encoding=encoding)
                self.output = OutputStream(output, compression=self.compression)

        self.streamer.write(self.output, points)

    def eof(self):
        if self.output is not None:
            self.streamer.eof(self.output)
            self.output.close()



import unittest

import mock

from dici import dici
from flume import *
from flume.exceptions import FlumineException
from robber import expect

class GithubPullsTest(unittest.TestCase):
    """
    verify that the third-party utilities for github pulls works
    """

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_repo_pulls(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.pulls.repo('whocares', 'repo', oauth='FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_repo_pulls(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.pulls.repo('whocares', 'repo', oauth='FAKETOKEN')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/repo/pulls',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_read_repo_pulls_with_parameters(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.pulls.repo('whocares',
                              'repo',
                              oauth='FAKETOKEN',
                              state='closed')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/repo/pulls?state=closed',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_pull_commits(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.pulls.commits('whocares', 'repo', 1, oauth='FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_pulls_commits(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.pulls.commits('whocares', 'repo', 1, oauth='FAKETOKEN')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/repo/pulls/1/commits',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])



"""
filter proc unittests
"""
import unittest

from robber import expect
from flume import *


class FilterTest(unittest.TestCase):

    def test_filter_on_a_numerical_field(self):
        results = []
        (
            emit(limit=10, start='2014-01-01', every='1s')
            | put(value=count())
            | filter('value <= 5')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2014-01-01T00:00:00.000Z', 'value': 1},
            {'time': '2014-01-01T00:00:01.000Z', 'value': 2},
            {'time': '2014-01-01T00:00:02.000Z', 'value': 3},
            {'time': '2014-01-01T00:00:03.000Z', 'value': 4},
            {'time': '2014-01-01T00:00:04.000Z', 'value': 5}
        ])

    def test_filter_on_a_string_field(self):
        results = []
        (
            emit(limit=3, start='2014-01-01', every='1s')
            | (put(foo='bar'), put(foo='baz'))
            | filter('foo == "bar"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2014-01-01T00:00:00.000Z', 'foo': 'bar'},
            {'time': '2014-01-01T00:00:01.000Z', 'foo': 'bar'},
            {'time': '2014-01-01T00:00:02.000Z', 'foo': 'bar'}
        ])

    def test_filter_on_a_boolean_field(self):
        results = []
        (
            emit(limit=3, start='2014-01-01', every='1s')
            | (put(foo=True), put(foo=False))
            | filter('foo')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2014-01-01T00:00:00.000Z', 'foo': True},
            {'time': '2014-01-01T00:00:01.000Z', 'foo': True},
            {'time': '2014-01-01T00:00:02.000Z', 'foo': True}
        ])



"""
timechart sink
"""

from flume import logger
from flume import sink
from flume.sinks.views import get_view


class timechart(sink):

    name = 'timechart'

    def __init__(self,
                 view,
                 series='name',
                 yvalue='value',
                 **kwargs):
        sink.__init__(self)
        self.series = series
        self.yvalue = yvalue
        self.view = get_view(view, **kwargs)

    def loop(self):
        series = {}

        had_data = False
        while self.running:
            points = self.pull()

            has_data = False
            for point in points:
                if 'time' not in point.keys():
                    logger.warn('skipping point without `time` field: %s' % point)
                    continue

                if self.series not in point:
                    logger.warn('unable to find series field "%s" in point' % self.series)
                    name = 'unknown'

                else:
                    name = point[self.series]

                if name not in series:
                    series[name] = []

                has_data = True
                had_data = True
                series[name].append((point.time, point[self.yvalue]))

            if has_data:
                self.view.render(self.name, series)

        if not had_data:
            logger.warn('no data to render')



"""
write sink
"""

from flume import adapters, sink


class write(sink):

    name = 'write'

    def __init__(self,
                 adapter,
                 batch=-1,
                 **kwargs):
        sink.__init__(self)
        cls = adapters.get_adapter(adapter, 'write')

        self.instance = cls(**kwargs)
        self.write = self.instance.write
        self.batch = batch

    def loop(self):
        buffered = []
        while self.running:
            points = self.pull()
            buffered += points

            if len(buffered) > 0:
                if self.batch == -1:
                    self.write(buffered)
                    buffered = []

                elif len(buffered) > self.batch:
                    self.write(buffered[:self.batch])
                    buffered = buffered[self.batch:]

        if len(buffered) > 0:
            self.write(buffered[:self.batch])

        self.instance.eof()



# -*- coding: utf-8 -*-
"""
intersect proc unittests
"""
import unittest

from robber import expect
from flume import *


class IntersectTest(unittest.TestCase):

    # test basic intersection properties from
    # https://en.wikipedia.org/wiki/Set_(mathematics)

    def test_idempotency(self):
        """
        A ∩ A = A
        """
        AnA = []

        A = [
            {'time': '2010-01-01T00:00:00.000Z', 'a': 0},
            {'time': '2010-01-01T00:01:00.000Z', 'a': 1},
            {'time': '2010-01-01T00:02:00.000Z', 'a': 2}
        ]

        (
            (emit(points=A), emit(points=A))
            | intersect()
            | memory(AnA)
        ).execute()

        expect(AnA).to.eq(A)

    def test_domination(self):
        """
        A ∩ ∅ = ∅
        """
        An0 = []

        A = [
            {'time': '2010-01-01T00:00:00.000Z', 'a': 0},
            {'time': '2010-01-01T00:01:00.000Z', 'a': 1},
            {'time': '2010-01-01T00:02:00.000Z', 'a': 2}
        ]

        (
            (emit(points=A), emit(points=[]))
            | intersect()
            | memory(An0)
        ).execute()

        expect(An0).to.eq([])

    def test_commutativity(self):
        """
        A ∩ B = B ∩ A
        """
        AnB = []
        BnA = []

        A = [
            {'time': '2010-01-01T00:00:00.000Z', 'a': 0},
            {'time': '2010-01-01T00:01:00.000Z', 'a': 1},
            {'time': '2010-01-01T00:02:00.000Z', 'a': 2}
        ]
        B = [
            {'time': '2010-01-01T00:01:00.000Z', 'a': 1},
            {'time': '2010-01-01T00:03:00.000Z', 'a': 2},
            {'time': '2010-01-01T00:05:00.000Z', 'a': 3}
        ]

        (
            (emit(points=A), emit(points=B))
            | intersect()
            | memory(AnB)
        ).execute()

        (
            (emit(points=B), emit(points=A))
            | intersect()
            | memory(BnA)
        ).execute()

        expect(AnB).to.eq(BnA)

    def test_associativity(self):
        """
        A ∩ (B ∩ C) = (A ∩ B) ∩ C.
        """
        ABC1 = []
        ABC2 = []

        A = [
            {'time': '2010-01-01T00:00:00.000Z', 'a': 0},
            {'time': '2010-01-01T00:01:00.000Z', 'a': 1},
            {'time': '2010-01-01T00:02:00.000Z', 'a': 2}
        ]
        B = [
            {'time': '2010-01-01T00:01:00.000Z', 'a': 1},
            {'time': '2010-01-01T00:03:00.000Z', 'a': 2},
            {'time': '2010-01-01T00:05:00.000Z', 'a': 3}
        ]
        C = [
            {'time': '2010-01-01T00:02:00.000Z', 'a': 3},
            {'time': '2010-01-01T00:04:00.000Z', 'a': 4},
            {'time': '2010-01-01T00:06:00.000Z', 'a': 5}
        ]

        (
            (
                emit(points=A),
                (emit(points=B), emit(points=C)) | intersect()
            )
            | intersect()
            | memory(ABC1)
        ).execute()

        (
            (
                (emit(points=A), emit(points=B)) | intersect(),
                emit(points=C)
            )
            | intersect()
            | memory(ABC2)
        ).execute()

        expect(ABC1).to.eq(ABC2)

    def test_empty_to_intersect_of_a_single_historical_stream(self):
        results = []
        (
            emit(limit=3, start='2016-01-01', every='1s')
            | ()
            | intersect()
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z'},
            {'time': '2016-01-01T00:00:01.000Z'},
            {'time': '2016-01-01T00:00:02.000Z'}
        ])

    def test_single_historical_stream(self):
        results = []
        (
            emit(limit=3, start='2000-01-01')
            | intersect()
            | memory(results)
        ).execute()
        expect(results).to.have.length(3)

    def test_single_live_stream(self):
        results = []
        (
            emit(limit=3, every='0.001s')
            | intersect()
            | memory(results)
        ).execute()
        expect(results).to.have.length(3)


    def test_empty_to_intersect_of_a_single_live_stream(self):
        results = []
        (
            emit(limit=3, every='0.001s')
            | ()
            | intersect()
            | memory(results)
        ).execute()
        expect(results).to.have.length(3)

    def test_intersect_two_historical_streams(self):
        results = []
        (
            emit(limit=3, start='2016-01-01', every='1s')
            | (
                put(foo='bar'),
                put(fizz='buzz')
            ) | intersect()
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'foo': 'bar', 'fizz': 'buzz'},
            {'time': '2016-01-01T00:00:01.000Z', 'foo': 'bar', 'fizz': 'buzz'},
            {'time': '2016-01-01T00:00:02.000Z', 'foo': 'bar', 'fizz': 'buzz'}
        ])

    def test_intersect_two_live_streams(self):
        results = []
        (
            emit(limit=3, every='0.001s')
            | (
                put(foo='bar'),
                put(fizz='buzz')
            ) | intersect()
            | memory(results)
        ).execute()
        expect(results).to.have.length(3)

    def test_intersect_on_a_non_time_field(self):
        results = []
        (
            ([
                emit(points=[
                    {'foo': 1, 'bar': 'baz'},
                    {'foo': 2, 'bar': 'baz'},
                    {'foo': 3, 'bar': 'baz'}
                ]),
                emit(points=[
                    {'foo': 1, 'baz': 'bar'},
                    {'foo': 2, 'baz': 'bar'},
                    {'foo': 3, 'baz': 'bar'}
                ])
            ])
            | intersect('foo')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1, 'bar': 'baz', 'baz': 'bar'},
            {'foo': 2, 'bar': 'baz', 'baz': 'bar'},
            {'foo': 3, 'bar': 'baz', 'baz': 'bar'}
        ])

    def test_intersect_on_multiple_fields(self):
        results = []
        (
            ([
                emit(points=[
                    {'foo': 1, 'bar': 'a', 'color': 'green'},
                    {'foo': 2, 'bar': 'b', 'color': 'blue'},
                ]),
                emit(points=[
                    {'foo': 2, 'bar': 'b', 'shape': 'circle'},
                    {'foo': 1, 'bar': 'a', 'shape': 'square'},
                ])
            ])
            | intersect('foo', 'bar')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1, 'bar': 'a', 'color': 'green', 'shape': 'square'},
            {'foo': 2, 'bar': 'b', 'color': 'blue', 'shape': 'circle'}
        ])



"""
seq proc unittests
"""
import unittest

from robber import expect
from flume import *


class SeqTest(unittest.TestCase):

    def test_seq_can_chain_two_historical_streams(self):
        results = []
        (
            seq(emit(limit=3, start='2016-01-01T00:00:00.000Z'),
                emit(limit=3, start='2016-01-01T00:00:03.000Z'))
            | put(value=count())
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'value': 1},
            {'time': '2016-01-01T00:00:01.000Z', 'value': 2},
            {'time': '2016-01-01T00:00:02.000Z', 'value': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'value': 4},
            {'time': '2016-01-01T00:00:04.000Z', 'value': 5},
            {'time': '2016-01-01T00:00:05.000Z', 'value': 6}
        ])


    def test_seq_can_chain_two_live_streams(self):
        results = []
        (
            seq(emit(limit=3, every='0.1s'),
                emit(limit=3, every='0.1s'))
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'count': 6}
        ])



from flume.procs.reduce.reducers.count import count
from flume.procs.reduce.reducers.iterate import iterate
from flume.procs.reduce.reducers.maximum import maximum
from flume.procs.reduce.reducers.minimum import minimum

from flume.procs.reduce.reducers.math import math
from flume.procs.reduce.reducers.date import date
from flume.procs.reduce.reducers.funcr import funcr



from flume.sinks.memory import memory
from flume.sinks.write import write

from flume.sinks.barchart import barchart
from flume.sinks.linechart import linechart
from flume.sinks.timechart import timechart



"""
common module

- defines various basic elements of what make up a flume pipeline
- handles all of the fancy logic for the dot(.) and pipe(|) notation when
  construction pipelines
"""
import bisect
import logging
import sys
import threading

import six

from flume import moment
from flume import logger
from flume import util

if util.IS_PY2:
    from Queue import Empty
    from Queue import Queue as queue

else:
    from queue import Queue as queue
    from queue import Empty

from flume.exceptions import FlumineException
from flume.point import Point


__SOURCES = {}
__PROCS = {}
__SINKS = {}


def build_node_func(which_node):
    """
    internally used method to attach new nodes to existing ones in the flume
    pipeline when registering such elements
    """

    def node_func(self, *args, **kwargs):
        """
        internal method responsible for hooking up the nodes in a flume
        pipeline
        """

        node_instance = which_node(*args, **kwargs)

        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[queue()])

        source = self.source if hasattr(self, 'source') else self
        node.init_node(node_instance,
                       inputs=self.outputs,
                       outputs=[queue()],
                       parent=self,
                       source=source)

        return node_instance

    return node_func

def register_source(source_class):
    """
    registers the source within flume and makes sure to setup any existing
    processors and sinks as chainable methods
    """
    if source_class.name in __SOURCES.keys():
        raise FlumineException('"%s" source already registered' % source_class.name)

    logger.debug('register_source %s', source_class.name)
    __SOURCES[source_class.name] = source_class

    for name in __PROCS.keys():
        setattr(source_class, name, build_node_func(__PROCS[name]))

    for sink_name in __SINKS.keys():
        setattr(source_class, sink_name, build_node_func(__SINKS[sink_name]))


def register_sink(sink_class):
    """
    registers the sink with flume and makes sure to hook itself up on any
    existing procs or sources as a chainable method
    """
    if sink_class.name in __SINKS.keys():
        raise FlumineException('"%s" sink already registered' % sink_class.name)

    logger.debug('register_sink %s', sink_class.name)
    __SINKS[sink_class.name] = sink_class

    for source in __SOURCES.values():
        setattr(source, sink_class.name, build_node_func(sink_class))

    for proc in __PROCS.values():
        setattr(proc, sink_class.name, build_node_func(sink_class))


def register_proc(proc_class):
    """
    registers the proc_class with flume and makes sure to hook any existing
    procs and sinks as chainable methods
    """
    if proc_class.name in __PROCS.keys():
        raise FlumineException('"%s" proc already registered' % proc_class.name)

    logger.debug('register_proc %s', proc_class.name)
    __PROCS[proc_class.name] = proc_class

    for source_class in __SOURCES.values():
        setattr(source_class, proc_class.name, build_node_func(proc_class))

    for other_proc_class in __PROCS.values():
        setattr(other_proc_class,
                proc_class.name,
                build_node_func(proc_class))
        setattr(proc_class,
                other_proc_class.name,
                build_node_func(other_proc_class))

    for sink_class in __SINKS.values():
        setattr(proc_class, sink.name, build_node_func(sink_class))

class node(object):
    """
    base flume node class
    """

    name = None

    EOF = Point(__eof=True)

    def __init__(self, *args, **kwargs):
        """
        the base node __init__ is just here to be record the arguments and
        keyword arguments used by each node
        """
        self.args = args
        self.kwargs = kwargs
        self.inited = True
        self.running = True
        self.exc_info = queue()
        self.inputs_index = None

    def init_node(self,
                  source=None,
                  parent=None,
                  inputs=None,
                  outputs=None):
        self.source = source
        self.inputs = inputs
        self.outputs = outputs
        self.parent = parent

    def pull(self, wait=True, timeout=None):
        """
        return a batch of points from the upstream procs and guarantee all
        points are sorted in order otherwise
        """
        times = []
        result = []

        if self.inputs is not None:

            if self.inputs_index is None:
                self.inputs_index = {}
                index = 0
                for input in self.inputs:
                    self.inputs_index[input] = index
                    index += 1

            for input in self.inputs:
                input_index = self.inputs_index[input]

                try:
                    points = input.get(wait)
                except Empty:
                    points = []

                for point in points:
                    point = Point(**point)

                    if point == self.EOF:
                        self.inputs.remove(input)
                        continue

                    point.__meta__.flume_path += ('.%s[%s]' % (self.name, input_index))

                    if 'time' not in point.keys():
                        result.append(point)

                    else:
                        index = bisect.bisect_left(times, point.time)
                        times.insert(index, point.time)
                        result.insert(index, point)

        if self.inputs is None or len(self.inputs) == 0:
            self.running = False

        logger.debug('%s pulling %s points', self, len(result))
        return result

    def push(self, points):
        """
        push the provided points to the outputs of this node so that other
        nodes downstream can receive it
        """

        if isinstance(points, Point):
            points = [points]

        logger.debug('%s pushing %s points', self, len(points))

        for point in points:
            for output in self.outputs:
                if point == self.EOF:
                    output.put([self.EOF])

                else:
                    output.put([Point(**point)])

    def __ror__(self, other):
        """

        """
        return node.__or__(splitter(flumes=other), self)

    def __or__(self, other):
        """
        enable the ability to use the bitwise or operator to emulate the pipe
        notation between flume nodes
        """
        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[queue()])

        source = self.source if hasattr(self, 'source') else self

        if isinstance(other, tuple):
            # split the stream!
            other = splitter(flumes=list(other))

        node.init_node(other,
                       inputs=self.outputs,
                       outputs=[queue()],
                       parent=self,
                       source=source)
        return other

    def loop(self):
        """
        every node should implement a loop method which is where they can handle
        the pushing/pulling of points
        """
        raise FlumineException('Each node should implement the loop method')

    def run(self):
        """
        node run method used by flume internally to manage thread execution
        """
        try:
            self.loop()
            self.exc_info.put(None)

        except:
            self.exc_info.put(sys.exc_info())

        finally:
            # must make sure to push EOF even if there was a failure
            self.push([self.EOF])

    def execute(self,
                wait=True,
                loglevel=logger.WARN):

        if 'inited' not in self.__dict__ or not self.inited:
            raise FlumineException('node.__init__ was never used')

        # XXX: special case of a single source with no output doing an execute()
        #      on it shouldn't fail
        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[])
        
        logger.setLogLevel(loglevel)

        # XXX: pooling here ?
        thread = threading.Thread(target=self.run)
        # Daemonize so that when we Ctrl+C the main program then all underlying
        # threads are instantly killed. Currently don't have any concern about
        # cleanly closing resources.
        thread.daemon = True
        thread.start()

        if self.parent:
            self.parent.execute(wait=wait, loglevel=loglevel)

        if wait:
            while thread.is_alive():
                # if you don't join with a timeout then you block the parent
                # until the child has completely finished and therefore can't
                # handle any signals in the parent (ie SIGINT)
                thread.join(1)

            exc_info = self.exc_info.get()

            if exc_info is not None:
                six.reraise(exc_info[0], exc_info[1], exc_info[2])

    def input_count(self):
        """
        returns how many inputs are currently hooked up to the current node
        """
        if self.parent.name == 'splitter':
            # XXX single point of hackery where we do something different for
            #     the split proc
            if len(self.parent.flumes) == 0:
                # since there's still the split input
                return 1

            else:
                return len(self.parent.flumes)
        else:
            return len(self.inputs)


class sink(node):
    """
    base sink class
    """

    def __init__(self, *args, **kwargs):
        node.__init__(self, *args, **kwargs)


class reducer(object):
    """
    base reducer class
    """

    def update(self, point):
        """
        update method called when there is a new point to be passed to the
        reducer
        """
        raise FlumineException('missing implementation for update method')

    def result(self):
        """
        reducer should return the reduced field value
        """
        raise FlumineException('missing implementation for result method')

class splitter(node):
    """
    # splitter

    The split processor is used internally to handle splitting the flume stream

    usage: ... | (put(...), filter(...), etc) | ...

    """

    name = 'splitter'

    def __init__(self,
                 flumes=None,
                 delay=None):
        node.__init__(self, flumes)
        self.flumes = flumes if flumes is not None else []

        if delay is not None:
            self.delay = moment.duration(delay)
        else:
            self.delay = moment.duration('5s')

        self.flume_outputs = []

    def loop(self):
        # self.input is None when split has no parent
        while self.running:
            points = self.pull()
            self.push(points)

    def execute(self,
                wait=True,
                loglevel=logging.ERROR):

        def find_root(flume):
            """
            figure out the root node of a flume pipeline
            """
            if hasattr(flume, 'parent') and \
               flume.parent is not None:
                return find_root(flume.parent)
            else:
                return flume

        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[])

        if len(self.flumes) != 0:

            forwarder_inputs = self.outputs
            self.outputs = []

            for flume in self.flumes:
                flume_input = queue()
                root = find_root(flume)
                root.parent = None
                root.inputs = [flume_input]

                # don't setup the outputs for a sink as it will never
                # push anything out XXX: could be a bit more elegant
                if not isinstance(flume, sink):
                    output = queue()
                    self.flume_outputs.append(output)
                    flume.outputs = [output]

                self.outputs.append(flume_input)

            from flume.procs import reorder
            forwarder = reorder(delay=self.delay)

            source = self.source if hasattr(self, 'source') else self
            node.init_node(forwarder,
                           inputs=self.flume_outputs,
                           outputs=forwarder_inputs,
                           parent=None,
                           source=source)

            forwarder.execute(wait=False,
                              loglevel=loglevel)

            # start underlying flumes
            for flume in self.flumes:
                flume.execute(wait=False,
                              loglevel=loglevel)

        # override default behavior to execute the underlying flume
        node.execute(self,
                     wait=False,
                     loglevel=loglevel)



"""
math reducers unittests
"""
import unittest

from datetime import datetime

from robber import expect
from flume import *


class DateReducerTest(unittest.TestCase):

    def test_date_year(self):
        results = []
        (
            emit(limit=1, start='2016-01-01')
            | put(year=date.year('time'))
            | keep('year')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'year': '2016'}
        ])

    def test_date_fullmonth(self):
        results = []
        (
            emit(limit=1, start='2016-01-01')
            | put(month=date.fullmonth('time'))
            | keep('month')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'month': datetime(2016, 1, 1).strftime('%B')}
        ])

    def test_date_month(self):
        results = []
        (
            emit(limit=1, start='2016-01-01')
            | put(month=date.month('time'))
            | keep('month')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'month': datetime(2016, 1, 1).strftime('%b')}
        ])

    def test_date_fullweekday(self):
        results = []
        (
            emit(limit=1, start='2016-01-01')
            | put(weekday=date.fullweekday('time'))
            | keep('weekday')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'weekday': datetime(2016, 1, 1).strftime('%A')}
        ])

    def test_date_weekday(self):
        results = []
        (
            emit(limit=1, start='2016-01-01')
            | put(weekday=date.weekday('time'))
            | keep('weekday')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'weekday': datetime(2016, 1, 1).strftime('%a')}
        ])

    def test_date_strftime(self):
        results = []
        (
            emit(limit=1, start='2016-01-01')
            | put(date=date.strftime('time', '%Y/%m/%d'))
            | keep('date')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'date': '2016/01/01'}
        ])



"""
math reducer - exposes the python mathematical functions as flume reducers
"""
import inspect

import math as realmath
from flume import reducer


# XXX: we could export this to be used to wrap any python function
def ReducerWrapper(func):

    class python_math_reducer(reducer):
        """
        a reducer that handles wrapping the provided function as a reducer
        over a single field
        """

        def __init__(self, fieldname):
            self.fieldname = fieldname
            self.value = None

        def update(self, point):
            self.value = func(point[self.fieldname])

        def result(self):
            return self.value

        def reset(self):
            self.value = None

    return python_math_reducer


class math(object):

    ceil = ReducerWrapper(realmath.ceil)
    fabs = ReducerWrapper(realmath.fabs)
    floor = ReducerWrapper(realmath.floor)
    factorial = ReducerWrapper(realmath.factorial)
    fmod = ReducerWrapper(realmath.fmod)
    trunc = ReducerWrapper(realmath.trunc)
    isinf = ReducerWrapper(realmath.isinf)
    isnan = ReducerWrapper(realmath.isnan)

    exp = ReducerWrapper(realmath.exp)
    expm1 = ReducerWrapper(realmath.expm1)
    log = ReducerWrapper(realmath.log)
    log1p = ReducerWrapper(realmath.log1p)
    log10 = ReducerWrapper(realmath.log10)
    pow = ReducerWrapper(realmath.pow)
    sqrt = ReducerWrapper(realmath.sqrt)

    acos = ReducerWrapper(realmath.acos)
    asin = ReducerWrapper(realmath.asin)
    atan = ReducerWrapper(realmath.atan)
    atan2 = ReducerWrapper(realmath.atan2)

    cos = ReducerWrapper(realmath.cos)
    sin = ReducerWrapper(realmath.sin)
    tan = ReducerWrapper(realmath.tan)

    degrees = ReducerWrapper(realmath.degrees)
    radians = ReducerWrapper(realmath.radians)



"""
elastic query module

 - handles the conversion of a filter to the ES query DSL
"""
import ast

from flume.exceptions import FlumineException


def to_es_query(node):
    """
    convert AST tree into an ES query
    """

    if isinstance(node, ast.Module):
        return {
            'filter': to_es_query(node.body[0])
        }

    if isinstance(node, list):
        # XXX: handle other elements of the list
        result = []
        nodes = node

        for node in nodes:
            result.append(to_es_query(node))

        return result

    if isinstance(node, ast.Expr):
        return to_es_query(node.value)

    if isinstance(node, ast.Num):
        return node.n

    if isinstance(node, ast.Str):
        return node.s

    if isinstance(node, ast.Compare):
        value = to_es_query(node.comparators[0])

        if isinstance(node.ops[0], ast.Eq):
            return {'term': {node.left.id: value}}

        if isinstance(node.ops[0], ast.NotEq):
            return {'not': {'term': {node.left.id: value}}}

        if isinstance(node.ops[0], ast.Lt):
            return {'range': {node.left.id: {'lt': value}}}

        if isinstance(node.ops[0], ast.Gt):
            return {'range': {node.left.id: {'gt': value}}}

        if isinstance(node.ops[0], ast.LtE):
            return {'range': {node.left.id: {'lte': value}}}

        if isinstance(node.ops[0], ast.GtE):
            return {'range': {node.left.id: {'gte': value}}}

    if isinstance(node, ast.BoolOp):
        if isinstance(node.op, ast.And):
            return {
                'bool': {
                    'must': to_es_query(node.values)
                }
            }

        elif isinstance(node.op, ast.Or):
            return {
                'bool': {
                    'should': to_es_query(node.values),
                    'minimum_should_match': 1
                }
            }

    if isinstance(node, ast.UnaryOp):
        if isinstance(node.op, ast.Not):
            return {
                'bool': {
                    'must_not': to_es_query(node.operand),
                }
            }

    raise Exception('unhandled %s' % node)

def filter_to_es_query(filter_expr):
    """
    take a filter expression such as `foo=="bar"` and convert it into
    the corresponding ES query
    """
    if filter_expr is None:
        return {'match_all': {}}
    else:
        try:
            tree = ast.parse(filter_expr)
            return {'constant_score': to_es_query(tree)}
        except Exception as exception:
            raise FlumineException('invalid filter expression: %s - %s' %
                                   (filter_expr, exception))



"""
testing utilities
"""

import sys

__IS_PY2 = sys.version[0] == '2'

class FakeIO(object):

    def __init__(self, data=''):
        self.data = data

    def close(self):
        pass

    def flush(self):
        pass

    def read(self, size):
        result = self.data[:size]
        self.data = self.data[size:]
        return result

    def readline(self, limit=-1):
        index = self.data.find('\n')

        if index == -1 and len(self.data) != 0:
            result = self.data
            self.data = ''
            return result

        result = self.data[:index + 1]
        self.data = self.data[index + 1:]
        return result

    def readlines(self):
        while True:
            line = self.readline()

            if not line:
                break

            yield line

    def write(self, data):
        self.data += data

    def getvalue(self):
        return self.data



"""
math reducers unittests
"""
import unittest

from robber import expect
from flume import *


class MathReducerTest(unittest.TestCase):

    def test_math_exp_log_reducers_work(self):
        results = []
        (
            emit(limit=3, every='0.001s')
            | put(count=count())
            | put(lvalue=math.log('count'))
            | put(value=math.exp('lvalue'))
            | keep('value')
            # handles rounding errors
            | put(value=math.floor('value'))
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'value': 1.0},
            {'value': 2.0},
            {'value': 3.0}
        ])




"""
gnuplot unit tests
"""
import unittest
import mock

from bunch import Bunch
from robber import expect

from flume.sinks.views import Gnuplot
from flume import moment


class GnuplotTest(unittest.TestCase):

    @mock.patch('tempfile.mkdtemp')
    @mock.patch('blessings.Terminal')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_render_a_simple_barchart_with_dumb_terminal(self,
                                                             mock_write,
                                                             mock_popen,
                                                             mock_terminal,
                                                             mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'
        mock_terminal.return_value = Bunch(width=100, height=50)

        view = Gnuplot()
        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 20), ('April', 5)]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('set terminal dumb size 100, 50\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            # default 0.8 boxwidth
            mock.call('set boxwidth 0.8\n'),
            # default fill style
            mock.call('set style fill solid\n'),
            mock.call('set style data histogram\n'),
            mock.call('set style histogram cluster gap 1\n'),
            mock.call('plot "/tmp/2016.dat" using 2:xtic(1) title "2016"\n')
        ])


    @mock.patch('tempfile.mkdtemp')
    @mock.patch('blessings.Terminal')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_render_a_simple_linechart_with_dumb_terminal(self,
                                                              mock_write,
                                                              mock_popen,
                                                              mock_terminal,
                                                              mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'
        mock_terminal.return_value = Bunch(width=100, height=50)

        view = Gnuplot()
        view.render('linechart', {
            'series1': [('0', 10), ('1', 20), ('2', 5)]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('set terminal dumb size 100, 50\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            mock.call('plot "/tmp/series1.dat" using 1:2 with linespoints title "series1"\n')
        ])


    @mock.patch('tempfile.mkdtemp')
    @mock.patch('blessings.Terminal')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_render_a_simple_timechart_with_dumb_terminal(self,
                                                              mock_write,
                                                              mock_popen,
                                                              mock_terminal,
                                                              mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'
        mock_terminal.return_value = Bunch(width=100, height=50)

        view = Gnuplot()
        view.render('timechart', {
            'series1': [
                (moment.date('2016-01-01T00:00:00.000Z'), 10),
                (moment.date('2016-01-01T00:00:01.000Z'), 20),
                (moment.date('2016-01-01T00:00:02.000Z'), 5)
            ]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('set terminal dumb size 100, 50\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            mock.call('set xdata time\n'),
            mock.call('set format x "%Y/%d/%m-%H:%M:%S"\n'),
            mock.call('set timefmt "%Y/%d/%m-%H:%M:%S"\n'),
            mock.call('plot "/tmp/series1.dat" using 1:2 with linespoints title "series1"\n')
        ])

    @mock.patch('tempfile.mkdtemp')
    @mock.patch('blessings.Terminal')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_call_render_multiple_times_for_a_simple_barchart_with_dumb_terminal(self,
                                                                                     mock_write,
                                                                                     mock_popen,
                                                                                     mock_terminal,
                                                                                     mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'
        mock_terminal.return_value = Bunch(width=100, height=50)

        view = Gnuplot()
        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 20), ('April', 5)]
        })

        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 10), ('April', 10)]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('set terminal dumb size 100, 50\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            # default 0.8 boxwidth
            mock.call('set boxwidth 0.8\n'),
            # default fill style
            mock.call('set style fill solid\n'),
            mock.call('set style data histogram\n'),
            mock.call('set style histogram cluster gap 1\n'),
            mock.call('plot "/tmp/2016.dat" using 2:xtic(1) title "2016"\n'),
            mock.call('exit\n'),
            mock.call('set terminal dumb size 100, 50\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:12.5]\n'),
            # default 0.8 boxwidth
            mock.call('set boxwidth 0.8\n'),
            # default fill style
            mock.call('set style fill solid\n'),
            mock.call('set style data histogram\n'),
            mock.call('set style histogram cluster gap 1\n'),
            mock.call('plot "/tmp/2016.dat" using 2:xtic(1) title "2016"\n')
        ])


    @mock.patch('tempfile.mkdtemp')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_render_a_simple_barchart_with_x11_terminal(self,
                                                            mock_write,
                                                            mock_popen,
                                                            mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'

        view = Gnuplot(terminal='x11')
        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 20), ('April', 5)]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('clear\n'),
            mock.call('set terminal x11 size 1280, 1024\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            # default 0.8 boxwidth
            mock.call('set boxwidth 0.8\n'),
            # default fill style
            mock.call('set style fill solid\n'),
            mock.call('set style data histogram\n'),
            mock.call('set style histogram cluster gap 1\n'),
            mock.call('plot "/tmp/2016.dat" using 2:xtic(1) title "2016"\n')
        ])


    @mock.patch('tempfile.mkdtemp')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_render_a_simple_linechart_with_x11_terminal(self,
                                                             mock_write,
                                                             mock_popen,
                                                             mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'

        view = Gnuplot(terminal='x11')
        view.render('linechart', {
            'series1': [('0', 10), ('1', 20), ('2', 5)]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('clear\n'),
            mock.call('set terminal x11 size 1280, 1024\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            mock.call('plot "/tmp/series1.dat" using 1:2 with linespoints title "series1"\n')
        ])


    @mock.patch('tempfile.mkdtemp')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_render_a_simple_timechart_with_x11_terminal(self,
                                                             mock_write,
                                                             mock_popen,
                                                             mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'

        view = Gnuplot(terminal='x11')
        view.render('timechart', {
            'series1': [
                (moment.date('2016-01-01T00:00:00.000Z'), 10),
                (moment.date('2016-01-01T00:00:01.000Z'), 20),
                (moment.date('2016-01-01T00:00:02.000Z'), 5)
            ]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('clear\n'),
            mock.call('set terminal x11 size 1280, 1024\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            mock.call('set xdata time\n'),
            mock.call('set format x "%Y/%d/%m-%H:%M:%S"\n'),
            mock.call('set timefmt "%Y/%d/%m-%H:%M:%S"\n'),
            mock.call('plot "/tmp/series1.dat" using 1:2 with linespoints title "series1"\n')
        ])

    @mock.patch('tempfile.mkdtemp')
    @mock.patch('subprocess.Popen')
    @mock.patch('flume.sinks.views.Gnuplot.write')
    def test_can_call_render_multiple_times_for_a_simple_barchart_with_x11_terminal(self,
                                                                                    mock_write,
                                                                                    mock_popen,
                                                                                    mock_mkdtemp):
        mock_mkdtemp.return_value = '/tmp'

        view = Gnuplot(terminal='x11')
        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 20), ('April', 5)]
        })

        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 10), ('April', 10)]
        })

        expect(mock_write.call_args_list).to.eq([
            mock.call('clear\n'),
            mock.call('set terminal x11 size 1280, 1024\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:25]\n'),
            # default 0.8 boxwidth
            mock.call('set boxwidth 0.8\n'),
            # default fill style
            mock.call('set style fill solid\n'),
            mock.call('set style data histogram\n'),
            mock.call('set style histogram cluster gap 1\n'),
            mock.call('plot "/tmp/2016.dat" using 2:xtic(1) title "2016"\n'),
            mock.call('clear\n'),
            mock.call('set terminal x11 size 1280, 1024\n'),
            # 25 comes from 20 * 1.25 we're doing 25% padding of the y value height
            mock.call('set yrange [0:12.5]\n'),
            # default 0.8 boxwidth
            mock.call('set boxwidth 0.8\n'),
            # default fill style
            mock.call('set style fill solid\n'),
            mock.call('set style data histogram\n'),
            mock.call('set style histogram cluster gap 1\n'),
            mock.call('plot "/tmp/2016.dat" using 2:xtic(1) title "2016"\n')
        ])



import unittest

from robber import expect
from flume import *
from flume.exceptions import FlumineException

from flume.core import register_sink, register_proc, register_source

class CoreTest(unittest.TestCase):
    """
    verifies basic flume flow behaviors
    """

    def test_failure_in_source_propagates_correctly(self):
        with self.assertRaisesRegexp(FlumineException, 'Unable to parse moment "-1"'):
            (
                emit(start=-1)
            ).execute()

    def test_failure_in_proc_propagates_correctly(self):
        class failme(reducer):

            def __init__(self):
                pass

            def update(self, point):
                raise Exception('failing on purpose')

        with self.assertRaisesRegexp(Exception, 'failing on purpose'):
            (
                emit(limit=10, start='2010-01-01')
                | put(foo=failme())
            ).execute()

    def test_failure_in_sink_propagates_correctly(self):
        with self.assertRaisesRegexp(Exception, 'Invalid URL \'bananas\''):
            (
                emit(limit=10, start='2010-01-01')
                | write('http', url='bananas')
            ).execute()

    def test_register_existing_source_fails(self):
        with self.assertRaisesRegexp(FlumineException, '"emit" source already registered'):
            register_source(emit)

    def test_register_existing_proc_fails(self):
        with self.assertRaisesRegexp(FlumineException, '"put" proc already registered'):
            register_proc(put)

    def test_register_existing_sink_fails(self):
        with self.assertRaisesRegexp(FlumineException, '"write" sink already registered'):
            register_sink(write)



"""
put proc unittests
"""
import unittest

from robber import expect
from flume import *


class PutTest(unittest.TestCase):

    def test_put_add_a_numerical_field(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo=1)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1},
            {'foo': 1}
        ])

    def test_put_add_a_nested_field(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo={
                'value': 1
            })
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': {'value': 1}},
            {'foo': {'value': 1}}
        ])

    def test_put_multiple_fields(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo=1, fizz='buzz')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1, 'fizz': 'buzz'},
            {'foo': 1, 'fizz': 'buzz'}
        ])

    def test_put_multiple_fields_with_multiple_puts(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo=1)
            | put(fizz='buzz')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1, 'fizz': 'buzz'},
            {'foo': 1, 'fizz': 'buzz'}
        ])

    def test_put_field_value_interpolations(self):
        results = []
        (
            emit(limit=3, start='2013-01-01')
            | put(count=count())
            | put(message='this is line #{count}')
            | remove('time', 'count')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'message': 'this is line #1'},
            {'message': 'this is line #2'},
            {'message': 'this is line #3'}
        ])

    def test_put_time_value_interpolations(self):
        results = []
        (
            emit(limit=3, start='2013-01-01')
            | put(count=count())
            | put(message='this happened at "{time}"')
            | remove('time', 'count')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'message': 'this happened at "2013-01-01T00:00:00.000Z"'},
            {'message': 'this happened at "2013-01-01T00:00:01.000Z"'},
            {'message': 'this happened at "2013-01-01T00:00:02.000Z"'}
        ])



"""
flume exceptions module
"""


class FlumineException(Exception):

    def __init__(self, message):
        # following may seem odd but necessary for this to work on both
        # python 2 and python 3
        Exception.__init__(self, message)
        self.message = message



"""
barchart sink
"""

from flume import logger
from flume import sink
from flume.sinks.views import get_view


class barchart(sink):

    name = 'barchart'

    def __init__(self,
                 view,
                 series='name',
                 category='time',
                 value='value',
                 **kwargs):
        sink.__init__(self)
        self.series = series
        self.category = category
        self.value = value
        self.view = get_view(view, **kwargs)

    def loop(self):
        series = {}

        while self.running:
            points = self.pull()

            for point in points:
                if self.series not in point:
                    logger.warn('unable to find series field "%s" in point' % self.series)
                    name = 'unknown'
                else:
                    name = point[self.series]

                if name not in series.keys():
                    series[name] = []

                series[name].append((point[self.category], point[self.value]))

            self.view.render('barchart', series)



import unittest

import mock

from dici import dici
from flume import *
from flume.exceptions import FlumineException
from robber import expect

class GithubIssuesTest(unittest.TestCase):
    """
    verify that the third-party utilities for github issues works
    """

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_user_issues(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.issues.user('FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_user_issues(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.user(oauth='FAKETOKEN')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/user/issues',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_read_user_issues_with_parameters(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.user(oauth='FAKETOKEN', filter='mentioned')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/user/issues?filter=mentioned',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_orgs_issues(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.issues.orgs('whocares', 'FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_orgs_issues(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (github.issues.orgs('whocares', 'FAKETOKEN') | memory(results)).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/orgs/whocares/issues',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_read_orgs_issues_with_parameters(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.orgs('whocares',
                               oauth='FAKETOKEN',
                               state='open')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/orgs/whocares/issues?state=open',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_repo_issues(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.issues.repo('whocares',
                               'somerepo',
                               oauth='FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_repo_issues(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (github.issues.repo('whocares',
                            'somerepo',
                            oauth='FAKETOKEN') | memory(results)).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/somerepo/issues',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_read_repo_issues_with_parameters(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.repo('whocares',
                            'somerepo',
                            oauth='FAKETOKEN',
                            sort='updated')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/somerepo/issues?sort=updated',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_issue(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.issues.issue('whocares',
                                'somerepo',
                                1,
                                oauth='FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_issue(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.issue('whocares', 'otherrepo', 1, oauth='FAKETOKEN')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/otherrepo/issues/1',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_handle_a_failure_from_issue_comments(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'oh shoot',
            'headers': [],
            'json': lambda: []
        })

        with self.assertRaisesRegexp(FlumineException, '400: oh shoot'):
            github.issues.comments('whocares',
                                   'somerepo',
                                   1,
                                   oauth='FAKETOKEN').execute()

    @mock.patch('requests.request')
    def test_can_read_issue_comments(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.comments('whocares', 'otherrepo', 1, oauth='FAKETOKEN')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/otherrepo/issues/1/comments',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])

    @mock.patch('requests.request')
    def test_can_read_issue_comments_with_parameters(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': [],
            'json': lambda: [{'created_at': '2016-01-01T00:00:00.000Z'}]
        })

        results = []
        (
            github.issues.comments('whocares',
                                   'otherrepo',
                                   1,
                                   oauth='FAKETOKEN',
                                   since='2016-01-01')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'time': '2016-01-01T00:00:00.000Z'}])
        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'https://api.github.com/repos/whocares/otherrepo/issues/1/comments?since=2016-01-01',
                      headers={
                          'Authorization': 'token FAKETOKEN'
                      })
        ])



"""
grok streamer
"""
import pygrok

from flume import Point, logger
from flume.adapters.streamers.base import Streamer
from flume.exceptions import FlumineException


class Grok(Streamer):

    def __init__(self,
                 pattern='%{GREEDYDATA:message}'):
        self.pattern = pattern

    def read(self, stream):
        for line in stream.readlines():
            matches = pygrok.grok_match(line, self.pattern)
            if matches is not None:
                yield Point(**matches)

            else:
                logger.warn('not matchined %s' % line)

    def write(self, stream, points):
        raise FlumineException('regex does not support writing data')



"""
emit source
"""

import time
import sched

from flume import moment, node, Point


class emit(node):

    name = 'emit'

    MAX_BUFFER = 1024

    def __init__(self,
                 limit=-1,
                 every='1s',
                 points=None,
                 start=None,
                 end=moment.end()):

        node.__init__(self,
                      limit=limit,
                      every=every,
                      points=points,
                      start=start,
                      end=end)
        self.limit = limit
        self.every = every
        self.start = moment.date(start) if start is not None else None
        self.end = moment.date(end)

        if points is None:
            self.points = None
        else:
            self.points = []

            for elem in points:
                point = Point(**elem)

                if 'time' in point:
                    point.time = moment.date(point.time)

                self.points.append(point)

    def loop(self):
        current_time = None

        if self.points is not None:
            self.push(self.points)

        else:
            count = 0

            if self.start is not None and self.start < moment.now():
                # emit the historical points first
                current_time = self.start
                buffered = []

                while current_time < moment.now():
                    count += 1

                    # done when we reach the limit
                    if self.limit != -1 and count > self.limit:
                        break

                    point = Point(time=current_time)
                    buffered.append(point)

                    if len(buffered) > emit.MAX_BUFFER:
                        self.push(buffered)
                        buffered = []

                    every = moment.duration(self.every, current_time)
                    current_time += every

                if len(buffered) > 0:
                    self.push(buffered)
                    buffered = []

            if moment.now() < self.end and count < self.limit:
                # continue into real time
                scheduler = sched.scheduler(time.time, time.sleep)
                every = moment.duration(self.every)
                every_seconds = every.total_seconds()

                if current_time is None:
                    if self.start is not None:
                        current_time = self.start
                    else:
                        current_time = moment.now()

                def push_point(current_time, count):
                    """
                    internal push point function
                    """
                    point = Point(time=current_time)
                    self.push([point])

                    count += 1
                    current_time += every

                    # done when we reach the limit
                    if self.limit != -1 and count >= self.limit:
                        return

                    # done when we've passed the `end`
                    if point.time > self.end:
                        return

                    scheduler.enter(every_seconds,
                                    1,
                                    push_point,
                                    (current_time, count))
            
                scheduler.enter(every_seconds,
                                1,
                                push_point,
                                (current_time, count))
                scheduler.run()



"""
filter processor
"""
import ast
import inspect

from flume import node

def islambda(something):
    return inspect.isfunction(something) and something.__name__ == '<lambda>'

class filter(node):

    name = 'filter'

    def __init__(self, filter_expression):
        node.__init__(self, filter_expression)
        self.filter = filter.parse_filter_expression(filter_expression)

    @classmethod
    def parse_filter_expression(cls, filter_expression):
        if isinstance(filter_expression, str):
            def wrap_eval(point):
                """
                wraps a filter expression encoded as a string 'a < "b"'
                """
                return eval(filter_expression, {}, point)

            return wrap_eval

        else:
            raise Exception('unsupported filter type "%s"' % type(filter_expression))

    def loop(self):

        while self.running:
            points = self.pull()
            points_out = []

            for point in points:
                if self.filter(point):
                    points_out.append(point)

            self.push(points_out)



"""
syslog flume extras that make reading from syslog files easy
"""
import os

from flume.procs import seq
from flume.sources import read

def __get_syslog_filenames(path):
    """
    internal utility to fetch the syslog files from the path provided knowing
    that the file starts with the prefix `syslog`
    """
    log_files = os.listdir(path)
    syslog_files = []
    for filename in log_files:
        if filename.startswith('syslog'):
            syslog_files.append(os.path.join(path, filename))

    return sorted(syslog_files, key=os.path.getmtime)


def syslog(logfile=None, path='/var/log/'):
    """
    syslog sink
    """
    if logfile is None:
        pipelines = []

        for logfile in __get_syslog_filenames(path):
            if logfile.endswith('.gz'):
                compression = 'gzip'

            else:
                compression = None

            pipelines.append(read('stdio',
                                  file=logfile,
                                  format='grok',
                                  pattern='%{SYSLOGLINE}',
                                  time='timestamp',
                                  compression=compression))

        return seq(*pipelines)
    else:
        if logfile.endswith('.gz'):
            compression = 'gzip'

        else:
            compression = None

        return read('stdio',
                    file=logfile,
                    format='grok',
                    pattern='%{SYSLOGLINE}',
                    time='timestamp',
                    compression=compression)



"""
views unit tests
"""
import unittest
import mock

from flume import *
from flume.adapters.adapter import adapter
from flume.sinks import views

from robber import expect


class ViewsTest(unittest.TestCase):

    def test_can_not_find_unregistered_view(self):
        with self.assertRaisesRegexp(FlumineException,
                                     'view with name "bananas", not found'):
            views.get_view('bananas')

    def test_can_not_register_existing_view(self):
        views.register_view('existing-view', object)
        with self.assertRaisesRegexp(FlumineException,
                                     'view with name "existing-view", already registered'):
            views.register_view('existing-view', object)

    def test_can_get_registered_view(self):
        class dummy(object):
            pass

        views.register_view('registered-view', dummy)
        result = views.get_view('registered-view')

        expect(result).to.be.instanceof(dummy)



"""
keep proc unittests
"""
import unittest

from robber import expect
from flume import *


class KeepTest(unittest.TestCase):

    def test_keep_a_numerical_field(self):
        results = []
        (
            emit(limit=3, every='0.01s')
            | put(value=count())
            | keep('value')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'value': 1},
            {'value': 2},
            {'value': 3}
        ])

    def test_keep_time_field(self):
        results = []
        (
            emit(limit=3, start='2016-01-01', every='1s')
            | put(value=count(), foo='bar')
            | keep('time')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z'},
            {'time': '2016-01-01T00:00:01.000Z'},
            {'time': '2016-01-01T00:00:02.000Z'}
        ])

    def test_keep_multiple_fields(self):
        results = []
        (
            emit(limit=3, start='2016-01-01', every='1s')
            | put(value=count(), foo='bar')
            | keep('time', 'value', 'foo')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'foo': 'bar', 'value': 1},
            {'time': '2016-01-01T00:00:01.000Z', 'foo': 'bar', 'value': 2},
            {'time': '2016-01-01T00:00:02.000Z', 'foo': 'bar', 'value': 3}
        ])



"""
jsonl streamer
"""
import json

from flume.adapters.streamers.base import Streamer
from flume import Point


class JSONL(Streamer):

    def __init__(self, pretty=False):
        self.pretty = pretty

    def read(self, stream):
        for line in stream.readlines():

            # skip blank lines
            if len(line.strip()) == 0:
                continue

            point = json.loads(line)
            yield Point(**point)

    def write(self, stream, points):
        for point in points:
            if self.pretty:
                # you can sort the keys in pretty jsonl format but indenting
                # would mean breaking the jsonl format
                string = json.dumps(point.json(), sort_keys=True)

            else:
                string = json.dumps(point.json())

            stream.write('%s\n' % string)



"""
gnuplot unit tests
"""
import unittest
import mock

from bunch import Bunch
from robber import expect

from flume.sinks.views import PyGal
from flume import moment

from test.unit.util import FakeIO

def bar():
    class Bar(object):
        kwargs = None
        series = {}

        def __init__(self, **kwargs):
            Bar.kwargs = kwargs
            Bar.series = {}

        def add(self, series_name, data):
            Bar.series[series_name] = data

        def render_to_png(self, filename):
            pass
        
        def render(self):
            pass

    return Bar

class PyGalTest(unittest.TestCase):

    @mock.patch('sys.stdout', new_callable=FakeIO)
    @mock.patch('tempfile.mkstemp')
    @mock.patch('pygal.Bar', new_callable=bar)
    def test_can_render_a_simple_barchart_to_png(self,
                                                 mock_bar,
                                                 mock_mkstemp,
                                                 mock_sys_stdout):
        mock_mkstemp.return_value = (None, '/tmp/barchart.png')

        view = PyGal(height=100, width=200)
        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 20), ('April', 5)]
        })

        # we spam stdout with the indication we're writing out a file
        expect(mock_sys_stdout.getvalue()).to.eq('rendering chart to "/tmp/barchart.png"\n')
        # internal initialization of pygal.Bar
        expect(mock_bar.kwargs).to.eq({
            'height': 100,
            'width': 200
        })
        # various calls to add series
        expect(mock_bar.series).to.eq({
            '2016': [10, 20, 5]
        })


    @mock.patch('sys.stdout', new_callable=FakeIO)
    @mock.patch('tempfile.mkstemp')
    @mock.patch('pygal.Bar', new_callable=bar)
    def test_can_render_a_simple_barchart_to_html(self,
                                                  mock_bar,
                                                  mock_mkstemp,
                                                  mock_sys_stdout):
        mock_mkstemp.return_value = (None, '/tmp/barchart.html')

        view = PyGal(height=100, width=200, format='html')
        view.render('barchart', {
            '2016': [('January', 10), ('Febuary', 20), ('April', 5)]
        })

        # we spam stdout with the indication we're writing out a file
        expect(mock_sys_stdout.getvalue()).to.eq('rendering chart to "/tmp/barchart.html"\n')
        # internal initialization of pygal.Bar
        expect(mock_bar.kwargs).to.eq({
            'height': 100,
            'width': 200
        })
        # various calls to add series
        expect(mock_bar.series).to.eq({
            '2016': [10, 20, 5]
        })



"""
streamer base class

"""
from flume.exceptions import FlumineException

class Streamer(object):

    def eof(self, stream):
        pass

    def read(self, stream):
        """
        read should return all of the data points present in the stream
        until the stream is empty
        """
        raise FlumineException('you must implement the read method')

    def write(self, stream, points):
        """
        write should write each of the points provided to the stream in
        the format of the output stream
        """
        raise FlumineException('you must implement the write method')



"""
flume command line utility """
import imp
import logging
import os
import warnings

import click

from flume import *

FLUMERC_LOCATIONS = [
    os.path.join(os.path.expanduser('~'), '.flumerc.py'),
    os.path.join(os.getcwd(), '.flumerc.py'),
]

import flume
VERSION = open(os.path.join(os.path.dirname(flume.__file__),
                            'VERSION')).read().strip()

@click.command()
@click.version_option(version=VERSION)
@click.option('--debug', '-d', count=True, help='debug mode')
@click.argument('program')
def main(program=None,
         debug=False):
    """
    simple command line entry point for executing flume programs
    """
    loglevel = logging.WARN

    if debug:
        loglevel = logging.DEBUG

    with warnings.catch_warnings():
        # lets ignore this silly warning
        message = 'Parent module \'\' not found while handling absolute import'
        warnings.filterwarnings('ignore', message=message)

        for flumerc in FLUMERC_LOCATIONS:
            if os.path.exists(flumerc):
                dirname = os.path.dirname(flumerc)
                fp, pathname, description = imp.find_module('.flumerc', [dirname])
                module = imp.load_module('.flumerc', fp, pathname, description)

                # load those .flumerc goodies into the global scope
                for thing in dir(module):
                    globals()[thing] = module.__dict__[thing]

    eval(program, globals(), locals()).execute(loglevel=loglevel)

if __name__ == '__main__':
    main()



"""
grok streamer unittests
"""
import unittest

from test.unit.util import FakeIO

from robber import expect

from flume import Point
from flume.adapters.streamers import get_streamer


class GrokStreamerTest(unittest.TestCase):

    def test_grok_streamer_can_read_an_empty_stream(self):
        streamer = get_streamer('grok')
        stream = FakeIO('')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([])

    def test_grok_streamer_can_read_a_single_point_with_default_pattern(self):
        streamer = get_streamer('grok')
        stream = FakeIO('just a simple line')

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(message='just a simple line')
        ])

    def test_grok_streamer_can_read_multiple_points_with_default_pattern(self):
        streamer = get_streamer('grok')
        stream = FakeIO('one log line\n' +
                        'another log line\n' +
                        'yet another log line\n')

        points = []
        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(message='one log line'),
            Point(message='another log line'),
            Point(message='yet another log line')
        ])

    def test_grok_streamer_can_handle_blank_lines_with_default_pattern(self):
        streamer = get_streamer('grok')
        stream = FakeIO('one log line\n' +
                        '\n' +
                        'another log line\n' +
                        '\n' +
                        'yet another log line\n')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(message='one log line'),
            Point(message=''),
            Point(message='another log line'),
            Point(message=''),
            Point(message='yet another log line')
        ])




"""
elastic adapter query conversion unittests
"""
import unittest

from robber import expect
from flume.adapters.elastic.query import filter_to_es_query
from flume.exceptions import FlumineException


class ElasticTest(unittest.TestCase):

    def test_default_filter_is_match_all(self):
        expect(filter_to_es_query(None)).to.eq({
            'match_all': {}
        })

    def test_invalid_assignment_query(self):
        try:
            expect(filter_to_es_query('foo="bar"'))
            raise Exception('previous statement should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('invalid filter expression: foo="bar"')

    def test_single_string_eq_query(self):
        expect(filter_to_es_query('foo=="bar"')).to.eq({
            'constant_score': {
                'filter': {
                    'term': {'foo': 'bar'}
                }
            }
        })

    def test_single_string_neq_query(self):
        expect(filter_to_es_query('foo!="bar"')).to.eq({
            'constant_score': {
                'filter': {
                    'not': {
                        'term': {'foo': 'bar'}
                    }
                }
            }
        })

    def test_single_term_integer_query(self):
        expect(filter_to_es_query('foo==2')).to.eq({
            'constant_score': {
                'filter': {
                    'term': {'foo': 2}
                }
            }
        })

    def test_single_and_query(self):
        expect(filter_to_es_query('foo==1 and fizz=="buzz"')).to.eq({
            'constant_score': {
                'filter': {
                    'bool': {
                        'must': [
                            {'term': {'foo': 1}},
                            {'term': {'fizz': 'buzz'}}
                        ]
                    }
                }
            }
        })

    def test_single_or_query(self):
        expect(filter_to_es_query('foo==1 or fizz=="buzz"')).to.eq({
            'constant_score': {
                'filter': {
                    'bool': {
                        'minimum_should_match': 1,
                        'should': [
                            {'term': {'foo': 1}},
                            {'term': {'fizz': 'buzz'}}
                        ]
                    }
                }
            }
        })

    def test_and_with_nested_or_query(self):
        query = 'foo==1 and (fizz=="buzz" or fizz=="bar")'

        expect(filter_to_es_query(query)).to.eq({
            'constant_score': {
                'filter': {
                    'bool': {
                        'must': [
                            {
                                'term': {'foo': 1}
                            },
                            {
                                'bool': {
                                    'minimum_should_match': 1,
                                    'should': [
                                        {'term': {'fizz': 'buzz'}},
                                        {'term': {'fizz': 'bar'}}
                                    ]
                                }
                            }
                        ]
                    }
                }
            }
        })

    def test_or_with_nested_and_query(self):
        query = 'foo==1 or (fizz=="buzz" and fizz=="bar")'
        expect(filter_to_es_query(query)).to.eq({
            'constant_score': {
                'filter': {
                    'bool': {
                        'minimum_should_match': 1,
                        'should': [
                            {
                                'term': {'foo': 1}
                            },
                            {
                                'bool': {
                                    'must': [
                                        {'term': {'fizz': 'buzz'}},
                                        {'term': {'fizz': 'bar'}}
                                    ]
                                }
                            }
                        ]
                    }
                }
            }
        })



"""
base view class
"""
from flume.exceptions import FlumineException

class base(object):

    def render(self, chart_type, data):
        """
        renders the provided data with the desired chart_type chart

        Arguments:
            chart_type - currently supported chart types are:
                linechart, timechart, barchart
            data - a dictionary of series names to lists of (x, y) values,
                like so:
                {
                    "series1": [(0, 0), (1, 1), ...],
                    "series2": [(0, 3), (1, 2), ...],
                    ...
                }

                linechart - the series names match to lines on the chart where
                            the x,y values are the pairs in the list.

                timechart - the series names match to lines on the chart where
                            the x,y values are the pairs in the list and x is
                            a datetime object.

                barchart - the series names are the names of the set of bars
                           that have a height of y for the x (may or may not be
                           a datetime object) multiple series represent multiple
                           bars for the same value of x.
        """
        raise FlumineException('implement me')



"""
github utilities
"""

from flume.util import IS_PY2

if IS_PY2:
    from urllib import urlencode

else:
    from urllib.parse import urlencode


def github_url(path, **kwargs):
    if len(kwargs) != 0:
        parameters = urlencode(kwargs)
        return 'https://api.github.com/%s?%s' % (path, parameters)

    else:
        return 'https://api.github.com/%s' % path



"""
http adapter unittests
"""

import threading
import time
import unittest

import json
import mock
import requests

from dici import dici
from robber import expect

from flume import *
from flume import moment

class HttpTest(unittest.TestCase):

    @mock.patch('requests.request')
    def test_http_read_with_expected_400_status(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'Bad Request',
            'headers': [],
            'json': lambda: []
        })

        (
            read('http',
                 url='http://localhost:8080/fail',
                 status=400)
        ).execute()

    @mock.patch('requests.request')
    def test_http_write_with_expected_400_status(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 400,
            'text': 'Bad Request',
            'headers': [],
            'json': lambda: []
        })

        (
            emit(limit=1)
            | write('http',
                 url='http://localhost:8080/fail',
                 status=400)
        ).execute()

    @mock.patch('requests.request')
    def test_http_read_with_bad_http_response(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 500,
            'text': 'Internal Server Error',
            'headers': []
        })

        try:
            (
                read('http',
                     url='http://localhost:8080/fail')
            ).execute()

            raise Exception('previous code should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('Internal Server Error')

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/fail', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_read_with_failure(self, mock_request):
        
        def raise_json_failure():
            raise ValueError('No JSON object could be decoded')

        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {
                'content-type': 'application/json'
            },
            'json': raise_json_failure
        })

        results = []
        try:
            (
                read('http',
                     url='http://localhost:8080/garbage')
                | memory(results)
            ).execute()

            raise Exception('previous code should have failed')
        except ValueError as exception:
            expect(str(exception)).to.eq('No JSON object could be decoded')
            expect(results).to.eq([])

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/garbage', headers=None)
        ])

    @mock.patch('requests.request')
    def test_empty_http_read(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {
                'content-type': 'application/json'
            },
            'json': lambda: json.loads('[]')
        })

        results = []
        (
            read('http',
                 url='http://localhost:8080/empty')
            | memory(results)
        ).execute()
        expect(results).to.eq([])

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/empty', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_read_a_few_historical_points(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {
                'content-type': 'application/json'
            },
            'json': lambda: [
                {'time': '2010-01-01T00:00:00.000Z'},
                {'time': '2010-01-01T00:00:01.000Z'},
                {'time': '2010-01-01T00:00:02.000Z'}
            ]
        })

        results = []
        (
            read('http',
                 url='http://localhost:8080/points?count=3&start=2010-01-01')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2010-01-01T00:00:00.000Z'},
            {'time': '2010-01-01T00:00:01.000Z'},
            {'time': '2010-01-01T00:00:02.000Z'}
        ])

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/points?count=3&start=2010-01-01', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_read_a_few_live_points(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {
                'content-type': 'application/json'
            },
            'json': lambda: [
                {'time': moment.now()},
                {'time': moment.now()},
                {'time': moment.now()},
                {'time': moment.now()},
                {'time': moment.now()}
            ]
        })

        results = []
        (
            read('http',
                 url='http://localhost:8080/points?count=5')
            | memory(results)
        ).execute()

        expect(results).to.have.length(5)

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/points?count=5', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_write_with_bad_http_response(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 500,
            'text': 'Internal Server Error',
        })

        results = []
        try:
            (
                emit(limit=1, start='2016-01-01')
                | write('http', url='http://localhost:8080/fail')
            ).execute()

            raise Exception('previous code should have failed')
        except FlumineException as exception:
            expect(exception.message).to.contain('Internal Server Error')

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET',
                      'http://localhost:8080/fail',
                      headers=None,
                      json=[{'time': '2016-01-01T00:00:00.000Z'}])
        ])

    @mock.patch('requests.request')
    def test_http_write_a_point_and_read_it_back(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {}
        })

        results = []
        (
            emit(limit=1, start='1970-01-01')
            | put(foo='bar')
            | write('http',
                    method='PUT',
                    url='http://localhost:8080/store?key=test1')
        ).execute()

        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {},
            'json': lambda: [{'time': '1970-01-01T00:00:00.000Z', 'foo': 'bar'}]
        })

        (
            read('http',
                 method='GET',
                 url='http://localhost:8080/store?key=test1')
            | keep('foo')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'foo': 'bar'}])

        expect(mock_request.call_args_list).to.eq([
            mock.call('PUT',
                      'http://localhost:8080/store?key=test1',
                      headers=None,
                      json=[{'time': '1970-01-01T00:00:00.000Z', 'foo': 'bar'}]),
            mock.call('GET', 'http://localhost:8080/store?key=test1', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_write_a_single_payload_and_read_them_back(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {}
        })

        results = []
        (
            emit(limit=3, start='1970-01-01')
            | put(count=count())
            | write('http',
                    method='PUT',
                    url='http://localhost:8080/store?key=test2',
                    batch=3)
        ).execute()

        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {},
            'json': lambda: [
                {'time': '1970-01-01T00:00:00.000Z', 'count': 1},
                {'time': '1970-01-01T00:00:01.000Z', 'count': 2},
                {'time': '1970-01-01T00:00:02.000Z', 'count': 3}
            ]
        })

        (
            read('http',
                 method='GET',
                 url='http://localhost:8080/store?key=test2')
            | keep('count')
            | memory(results)
        ).execute()

        # if we wrote out multiple paylaods then the end result would be that
        # we would only be able to retrieve the last point
        expect(results).to.eq([
            {'count': 1},
            {'count': 2},
            {'count': 3}
        ])

        expect(mock_request.call_args_list).to.eq([
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json=[
                          {'time': '1970-01-01T00:00:00.000Z', 'count': 1},
                          {'time': '1970-01-01T00:00:01.000Z', 'count': 2},
                          {'time': '1970-01-01T00:00:02.000Z', 'count': 3}
                      ]),
            mock.call('GET', 'http://localhost:8080/store?key=test2', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_write_multiple_payloads_and_read_them_back(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {}
        })

        results = []
        (
            emit(limit=3, start='1970-01-01')
            | put(count=count())
            | write('http',
                    method='PUT',
                    url='http://localhost:8080/store?key=test2',
                    batch=1)
        ).execute()

        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {},
            'json': lambda: [
                {'time': '1970-01-01T00:00:00.000Z', 'count': 3}
            ]
        })

        (
            read('http',
                 method='GET',
                 url='http://localhost:8080/store?key=test2')
            | keep('count')
            | memory(results)
        ).execute()

        # because the payload is of a single point we only end up storing the
        # last point that was pushed
        expect(results).to.eq([
            {'count': 3}
        ])

        expect(mock_request.call_args_list).to.eq([
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json=[{'time': '1970-01-01T00:00:00.000Z', 'count': 1}]),
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json=[{'time': '1970-01-01T00:00:01.000Z', 'count': 2}]),
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json=[{'time': '1970-01-01T00:00:02.000Z', 'count': 3}]),
            mock.call('GET', 'http://localhost:8080/store?key=test2', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_write_single_points(self, mock_request):
        mock_request.return_value = dici(**{
            'status_code': 200,
            'headers': {}
        })

        (
            emit(limit=3, start='1970-01-01')
            | put(count=count())
            | write('http',
                    method='PUT',
                    url='http://localhost:8080/store?key=test2',
                    array=False)
        ).execute()

        expect(mock_request.call_args_list).to.eq([
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json={'time': '1970-01-01T00:00:00.000Z', 'count': 1}),
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json={'time': '1970-01-01T00:00:01.000Z', 'count': 2}),
            mock.call('PUT',
                      'http://localhost:8080/store?key=test2',
                      headers=None,
                      json={'time': '1970-01-01T00:00:02.000Z', 'count': 3}),
        ])

    @mock.patch('requests.request')
    def test_http_read_can_follow_link_header(self, mock_request):
        points = [
            {'time': '2016-01-01T00:00:00.000Z'},
            {'time': '2016-01-01T00:00:01.000Z'}
        ]
        mock_request.side_effect = [
            dici(**{
                'status_code': 200,
                'text': 'OK',
                'headers': {
                    'link': '<http://localhost:8080/?count=2&page=2>; rel="next"'
                },
                'links': {
                    'next': {
                        'url': 'http://localhost:8080/points?count=2&page=2'
                    }
                },
                'json': lambda: points
            }),
            dici(**{
                'status_code': 200,
                'text': 'OK',
                'headers': {
                    'link': '<http://localhost:8080/?count=2&page=1>; rel="next"'
                },
                'links': {
                    'next': {
                        'url': 'http://localhost:8080/points?count=2&page=1'
                    }
                },
                'json': lambda: points
            }),
            dici(**{
                'status_code': 200,
                'text': 'OK',
                'links': {},
                'headers': {},
                'json': lambda: points
            }),
        ]

        results = []
        (
            read('http',
                 method='GET',
                 url='http://localhost:8080/points?count=2&page=3')
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'count': 6}])

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/points?count=2&page=3', headers=None),
            mock.call('GET', 'http://localhost:8080/points?count=2&page=2', headers=None),
            mock.call('GET', 'http://localhost:8080/points?count=2&page=1', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_read_can_not_follow_link_header(self, mock_request):
        points = [
            {'time': '2016-01-01T00:00:00.000Z'},
            {'time': '2016-01-01T00:00:01.000Z'}
        ]
        mock_request.side_effect = [
            dici(**{
                'status_code': 200,
                'text': 'OK',
                'headers': {
                    'link': '<http://localhost:8080/?page=2&count=2>; rel="next"'
                },
                'links': {
                    'next': {
                        'url': 'http://localhost:8080/points?page=2&count=2'
                    }
                },
                'json': lambda: points
            }),
            dici(**{
                'status_code': 200,
                'text': 'OK',
                'headers': {
                    'link': '<http://localhost:8080/?page=1&count=2>; rel="next"'
                },
                'links': {
                    'next': {
                        'url': 'http://localhost:8080/points?page=1&count=2'
                    }
                },
                'json': lambda: points
            }),
            dici(**{
                'status_code': 200,
                'text': 'OK',
                'links': {},
                'headers': {},
                'json': lambda: points
            }),
        ]

        results = []
        (
            read('http',
                 method='GET',
                 url='http://localhost:8080/points?count=2&page=3',
                 follow_link=False)
            | reduce(count=count())
            | keep('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'count': 2}])

        expect(mock_request.call_args_list).to.eq([
            mock.call('GET', 'http://localhost:8080/points?count=2&page=3', headers=None)
        ])

    @mock.patch('requests.request')
    def test_http_read_can_handle_grok_format_corectly(self, mock_request):
        with open('examples/grok/syslog') as syslog:
            mock_request.return_value = dici(**{
                'status_code': 200,
                'iter_content': lambda chunk_size: syslog.readlines(),
                'headers': []
            })

            results = []
            (
                read('http',
                     method='GET',
                     url='http://localhost:8080/syslog',
                     format='grok',
                     pattern='%{SYSLOGLINE}')
                | reduce(count=count())
                | keep('count')
                | memory(results)
            ).execute()

            expect(results).to.eq([{'count': 274}])
            expect(mock_request.call_args_list).to.eq([
                mock.call('GET', 'http://localhost:8080/syslog', headers=None, stream=True)
            ])



"""
json streamer unittests
"""
import unittest

from robber import expect

from flume.adapters.streamers import get_streamer
from flume import Point
from test.unit.util import FakeIO


class JSONStreamerTest(unittest.TestCase):

    def test_json_streamer_can_read_an_empty_stream(self):
        streamer = get_streamer('json')
        stream = FakeIO('[]')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([])

    def test_json_streamer_can_read_a_single_point(self):
        streamer = get_streamer('json')
        stream = FakeIO('[{"foo": "bar"}]')
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([Point(foo='bar')])

    def test_json_streamer_can_read_multiple_points(self):
        streamer = get_streamer('json')
        stream = FakeIO("""[{"time": "2016-01-01T00:00:00.000Z", "foo": 1},
        {"time": "2016-01-01T00:00:01.000Z", "foo": 2},
        {"time": "2016-01-01T00:00:02.000Z", "foo": 3}
        ]""")
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z', foo=1),
            Point(time='2016-01-01T00:00:01.000Z', foo=2),
            Point(time='2016-01-01T00:00:02.000Z', foo=3),
        ])

    def test_json_streamer_can_handle_blank_lines(self):
        streamer = get_streamer('json')
        stream = FakeIO("""[
        {"time": "2016-01-01T00:00:00.000Z", "foo": 1},
        {"time": "2016-01-01T00:00:01.000Z", "foo": 2}
        ]""")
        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z', foo=1),
            Point(time='2016-01-01T00:00:01.000Z', foo=2)
        ])

    def test_json_streamer_can_write_an_empty_stream(self):
        streamer = get_streamer('json')
        stream = FakeIO('')

        streamer.write(stream, [])
        streamer.eof(stream)
        stream = FakeIO(stream.getvalue())

        points = []
        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([])

    def test_json_streamer_write_a_single_point(self):
        streamer = get_streamer('json')
        stream = FakeIO('')

        streamer.write(stream, [Point(foo='bar')])
        streamer.eof(stream)
        stream = FakeIO(stream.getvalue())

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([Point(foo='bar')])

    def test_json_streamer_write_a_single_pretty_point(self):
        streamer = get_streamer('json', pretty=True)
        stream = FakeIO('')

        streamer.write(stream, [Point(foo='bar')])
        streamer.eof(stream)
        stream = FakeIO(stream.getvalue())

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([Point(foo='bar')])

    def test_json_streamer_can_write_multiple_points(self):
        streamer = get_streamer('json')
        stream = FakeIO('')

        streamer.write(stream, [
            Point(time='2016-01-01T00:00:00.000Z', foo=1),
            Point(time='2016-01-01T00:00:01.000Z', foo=2),
            Point(time='2016-01-01T00:00:02.000Z', foo=3)
        ])
        streamer.eof(stream)
        stream = FakeIO(stream.getvalue())

        points = []

        for point in streamer.read(stream):
            points.append(point)

        expect(points).to.eq([
            Point(time='2016-01-01T00:00:00.000Z', foo=1),
            Point(time='2016-01-01T00:00:01.000Z', foo=2),
            Point(time='2016-01-01T00:00:02.000Z', foo=3)
        ])



"""
elastic adapter query conversion unittests
"""
import time
import unittest

import requests

from docker import Client
from robber import expect

from flume import *


class ElasticTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cli = Client()
        host_config = cls.cli.create_host_config(port_bindings={
            9200:9200
        })
        cls.container = cls.cli.create_container(image='elasticsearch:2.3.3',
                                                 ports=[9200],
                                                 host_config=host_config)
        cls.cli.start(container=cls.container.get('Id'))

        # wait for elasticsearch to be running
        response = None
        start = time.time()
        while (response is None or response.status_code != 200) and \
              (time.time() - start) < 30000: # timeout after 30s
            try:
                response = requests.get('http://localhost:9200/_cluster/health')
            except requests.ConnectionError:
                pass
            time.sleep(1)

        if response is None or response.status_code != 200:
            raise Exception('unable to bring up elasticsearch v2.3.3')

        # create 10 points in a test index
        (
            emit(limit=10, start='2016-01-01')
            | put(string=iterate(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']),
                  integer=count(),
                  float=math.ceil('integer'))
            | write('elastic', index='test_data')
        ).execute()

        # elasticsearch commit delay
        time.sleep(1)

    @classmethod
    def tearDownClass(cls):
        cls.cli.stop(container=cls.container.get('Id'))
        cls.cli.wait(container=cls.container.get('Id'))

    def test_elastic_read_fails_on_inexistent_index(self):
        try:
            (
                read('elastic',
                     index='bananas',
                     filter='foo=="bananas"')
            ).execute()
        except Exception as exception:
            expect(str(exception)).to.contain('no such index')

    def test_elastic_read_empty(self):
        results = []

        (
            read('elastic',
                 index='test_data',
                 filter='foo=="bananas"')
            | memory(results)
        ).execute()

        expect(results).to.eq([])

    def test_elastic_read_with_small_batch(self):
        results = []

        (
            read('elastic',
                 index='test_data',
                 filter='integer > 0',
                 batch=2)
            | memory(results)
        ).execute()
        expect(results).to.have.length(10)

    def test_elastic_read_with_filter_eq_string(self):
        results = []
        (
            read('elastic', index='test_data', filter='string == "h"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0}
        ])

    def test_elastic_read_with_filter_eq_integer(self):
        results = []
        (
            read('elastic', index='test_data', filter='integer == 8')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0}
        ])

    def test_elastic_read_with_filter_eq_float(self):
        results = []
        (
            read('elastic', index='test_data', filter='float == 8.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0}
        ])

    def test_elastic_read_with_filter_neq_string(self):
        results = []
        (
            read('elastic', index='test_data', filter='string != "h"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0},
            {'time': '2016-01-01T00:00:03.000Z', 'string': 'd', 'integer': 4, 'float': 4.0},
            {'time': '2016-01-01T00:00:04.000Z', 'string': 'e', 'integer': 5, 'float': 5.0},
            {'time': '2016-01-01T00:00:05.000Z', 'string': 'f', 'integer': 6, 'float': 6.0},
            {'time': '2016-01-01T00:00:06.000Z', 'string': 'g', 'integer': 7, 'float': 7.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_neq_integer(self):
        results = []
        (
            read('elastic', index='test_data', filter='integer != 8')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0},
            {'time': '2016-01-01T00:00:03.000Z', 'string': 'd', 'integer': 4, 'float': 4.0},
            {'time': '2016-01-01T00:00:04.000Z', 'string': 'e', 'integer': 5, 'float': 5.0},
            {'time': '2016-01-01T00:00:05.000Z', 'string': 'f', 'integer': 6, 'float': 6.0},
            {'time': '2016-01-01T00:00:06.000Z', 'string': 'g', 'integer': 7, 'float': 7.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_neq_float(self):
        results = []
        (
            read('elastic', index='test_data', filter='float != 8.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0},
            {'time': '2016-01-01T00:00:03.000Z', 'string': 'd', 'integer': 4, 'float': 4.0},
            {'time': '2016-01-01T00:00:04.000Z', 'string': 'e', 'integer': 5, 'float': 5.0},
            {'time': '2016-01-01T00:00:05.000Z', 'string': 'f', 'integer': 6, 'float': 6.0},
            {'time': '2016-01-01T00:00:06.000Z', 'string': 'g', 'integer': 7, 'float': 7.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_gt_string(self):
        results = []
        (
            read('elastic', index='test_data', filter='string > "g"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_gt_integer(self):
        results = []
        (
            read('elastic', index='test_data', filter='integer > 7')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_gt_float(self):
        results = []
        (
            read('elastic', index='test_data', filter='float > 7.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_gte_string(self):
        results = []
        (
            read('elastic', index='test_data', filter='string >= "h"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_gte_integer(self):
        results = []
        (
            read('elastic', index='test_data', filter='integer >= 8')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_gte_float(self):
        results = []
        (
            read('elastic', index='test_data', filter='float >= 8.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:07.000Z', 'string': 'h', 'integer': 8, 'float': 8.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])

    def test_elastic_read_with_filter_lt_string(self):
        results = []
        (
            read('elastic', index='test_data', filter='string < "d"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_filter_lt_integer(self):
        results = []
        (
            read('elastic', index='test_data', filter='integer < 4')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_filter_lt_float(self):
        results = []
        (
            read('elastic', index='test_data', filter='float < 4.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_filter_lte_string(self):
        results = []
        (
            read('elastic', index='test_data', filter='string <= "c"')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_filter_lte_integer(self):
        results = []
        (
            read('elastic', index='test_data', filter='integer <= 3')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_filter_lte_float(self):
        results = []
        (
            read('elastic', index='test_data', filter='float <= 3.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_and_filter(self):
        results = []
        (
            read('elastic', index='test_data', filter='string == "a" and float == 1.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0}
        ])

    def test_elastic_read_with_or_filter(self):
        results = []
        (
            read('elastic', index='test_data', filter='string == "a" or float == 3.0')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0}
        ])

    def test_elastic_read_with_not_filter(self):
        results = []
        (
            read('elastic', index='test_data', filter='not(string == "h")')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'string': 'a', 'integer': 1, 'float': 1.0},
            {'time': '2016-01-01T00:00:01.000Z', 'string': 'b', 'integer': 2, 'float': 2.0},
            {'time': '2016-01-01T00:00:02.000Z', 'string': 'c', 'integer': 3, 'float': 3.0},
            {'time': '2016-01-01T00:00:03.000Z', 'string': 'd', 'integer': 4, 'float': 4.0},
            {'time': '2016-01-01T00:00:04.000Z', 'string': 'e', 'integer': 5, 'float': 5.0},
            {'time': '2016-01-01T00:00:05.000Z', 'string': 'f', 'integer': 6, 'float': 6.0},
            {'time': '2016-01-01T00:00:06.000Z', 'string': 'g', 'integer': 7, 'float': 7.0},
            {'time': '2016-01-01T00:00:08.000Z', 'string': 'i', 'integer': 9, 'float': 9.0},
            {'time': '2016-01-01T00:00:09.000Z', 'string': 'j', 'integer': 10, 'float': 10.0}
        ])
